import types
from typing import Any, List, Tuple

import cv2
import numpy
from PIL.ImageTransform import Transform
from numpy import random

__all__ = [
    "CV2Compose",
    "Lambda",
    "ConvertFromInts",
    "SubtractMeans",
    "CV2ToAbsoluteCoords",
    "CV2ToPercentCoords",
    "CV2Resize",
    "CV2RandomSaturation",
    "CV2RandomHue",
    "CV2RandomLightingNoise",
    "CV2ConvertColor",
    "CV2RandomContrast",
    "CV2RandomBrightness",
    "CV2RandomSampleCrop",
    "CV2Expand",
    "CV2RandomMirror",
    "CV2SwapChannels",
    "CV2PhotometricDistort",
]

from draugr.opencv_utilities.bounding_boxes.evaluation import (
    jaccard_overlap_numpy,
    remove_null_boxes,
)
from warg import TripleNumber


class CV2Compose(object):
    """Composes several augmentations together.
    Args:
    transforms (List[Transform]): list of transforms to compose.
    Example:
    >>> augmentations.Compose([
    >>>     transforms.CenterCrop(10),
    >>>     transforms.ToTensor(),
    >>> ])"""

    def __init__(self, transforms: List[Transform]):
        self._transforms = transforms

    def __call__(self, img, boxes=None, labels=None) -> Tuple:
        for transform in self._transforms:
            img, boxes, labels = transform(img, boxes, labels)
            if boxes is not None:
                boxes, labels = remove_null_boxes(boxes, labels)
        return img, boxes, labels


class Lambda(object):
    """Applies a lambda as a transform."""

    def __init__(self, lambd: callable):
        assert isinstance(lambd, types.LambdaType)
        self.lambd = lambd

    def __call__(self, img: Any, boxes: Any = None, labels: Any = None):
        return self.lambd(img, boxes, labels)


class ConvertFromInts(object):
    def __call__(self, image: Any, boxes: Any = None, labels: Any = None) -> Tuple:
        return image.astype(numpy.float32), boxes, labels


class SubtractMeans(object):
    """ """

    def __init__(self, mean: Tuple):
        self.mean = numpy.array(mean, dtype=numpy.float32)

    def __call__(self, image: Any, boxes: Any = None, labels: Any = None) -> Tuple:
        image = image.astype(numpy.float32)
        image -= self.mean
        return image.astype(numpy.float32), boxes, labels


class CV2ToAbsoluteCoords(object):
    def __call__(self, image: Any, boxes: Any = None, labels: Any = None) -> Tuple:
        height, width, channels = image.shape
        boxes[:, 0] *= width
        boxes[:, 2] *= width
        boxes[:, 1] *= height
        boxes[:, 3] *= height

        return image, boxes, labels


class CV2ToPercentCoords(object):
    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        height, width, channels = image.shape
        boxes[:, 0] /= width
        boxes[:, 2] /= width
        boxes[:, 1] /= height
        boxes[:, 3] /= height

        return image, boxes, labels


class CV2Resize(object):
    """ """

    def __init__(self, size: int = 300):
        self._size = size

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        image = cv2.resize(image, (self._size, self._size))
        return image, boxes, labels


class CV2RandomSaturation(object):
    """ """

    def __init__(self, lower: float = 0.5, upper: float = 1.5):
        self.lower = lower
        self.upper = upper
        assert self.upper >= self.lower, "contrast upper must be >= lower."
        assert self.lower >= 0, "contrast lower must be non-negative."

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        if random.randint(2):
            image[..., 1] *= random.uniform(self.lower, self.upper)

        return image, boxes, labels


class CV2RandomHue(object):
    """ """

    def __init__(self, delta: float = 18.0):
        assert delta >= 0.0 and delta <= 360.0
        self.delta = delta

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        if random.randint(2):
            image[..., 0] += random.uniform(-self.delta, self.delta)
            image[..., 0][image[..., 0] > 360.0] -= 360.0
            image[..., 0][image[..., 0] < 0.0] += 360.0
        return image, boxes, labels


class CV2RandomLightingNoise(object):
    """ """

    def __init__(self):
        self.perms = ((0, 1, 2), (0, 2, 1), (1, 0, 2), (1, 2, 0), (2, 0, 1), (2, 1, 0))

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        if random.randint(2):
            swap = self.perms[random.randint(len(self.perms))]
            shuffle = CV2SwapChannels(swap)  # shuffle channels
            image = shuffle(image)
        return image, boxes, labels


class CV2ConvertColor(object):
    """ """

    def __init__(self, current, transform):
        self.transform = transform
        self.current = current

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        if self.current == "BGR" and self.transform == "HSV":
            image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        elif self.current == "RGB" and self.transform == "HSV":
            image = cv2.cvtColor(image, cv2.COLOR_RGB2HSV)
        elif self.current == "BGR" and self.transform == "RGB":
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
        elif self.current == "HSV" and self.transform == "BGR":
            image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)
        elif self.current == "HSV" and self.transform == "RGB":
            image = cv2.cvtColor(image, cv2.COLOR_HSV2RGB)
        else:
            raise NotImplementedError
        return image, boxes, labels


class CV2RandomContrast(object):
    """ """

    def __init__(self, lower: float = 0.5, upper: float = 1.5):
        self.lower = lower
        self.upper = upper
        assert self.upper >= self.lower, "contrast upper must be >= lower."
        assert self.lower >= 0, "contrast lower must be non-negative."

    # expects float image
    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        if random.randint(2):
            alpha = random.uniform(self.lower, self.upper)
            image *= alpha
        return image, boxes, labels


class CV2RandomBrightness(object):
    """ """

    def __init__(self, delta: float = 32):
        assert delta >= 0.0
        assert delta <= 255.0
        self.delta = delta

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        if random.randint(2):
            delta = random.uniform(-self.delta, self.delta)
            image += delta
        return image, boxes, labels


class CV2RandomSampleCrop(object):
    """Crop
    Arguments:
    img (Image): the image being input during training
    boxes (Tensor): the original bounding boxes in pt form
    labels (Tensor): the class labels for each bbox
    mode (float tuple): the min and max jaccard overlaps
    Return:
    (img, boxes, classes)
    img (Image): the cropped image
    boxes (Tensor): the adjusted bounding boxes in pt form
    labels (Tensor): the class labels for each bbox"""

    def __init__(self):
        self.sample_options = (
            # using entire original input image
            None,
            # sample a patch s.t. MIN jaccard w/ obj in .1,.3,.4,.7,.9
            (0.1, None),
            (0.3, None),
            (0.7, None),
            (0.9, None),
            # randomly sample a patch
            (None, None),
        )

    def __call__(
        self,
        image: numpy.ndarray,
        boxes: numpy.ndarray = None,
        labels: numpy.ndarray = None,
    ) -> Tuple:
        # guard against no boxes
        if boxes is not None and boxes.shape[0] == 0:
            return image, boxes, labels
        height, width, _ = image.shape
        while True:
            # randomly choose a mode
            mode = random.choice(self.sample_options)
            if mode is None:
                return image, boxes, labels

            min_iou, max_iou = mode
            if min_iou is None:
                min_iou = float("-inf")
            if max_iou is None:
                max_iou = float("inf")

            # max trails (50)
            for _ in range(50):
                current_image = image

                w = random.uniform(0.3 * width, width)
                h = random.uniform(0.3 * height, height)

                # aspect ratio constraint b/t .5 & 2
                if h / w < 0.5 or h / w > 2:
                    continue

                left = random.uniform(width - w)
                top = random.uniform(height - h)

                # convert to integer rect x1,y1,x2,y2
                rect = numpy.array([int(left), int(top), int(left + w), int(top + h)])

                # calculate IoU (jaccard overlap) b/t the cropped and gt boxes
                overlap = jaccard_overlap_numpy(boxes, rect)

                # is min and max overlap constraint satisfied? if not try again
                if overlap.max() < min_iou or overlap.min() > max_iou:
                    continue

                # cut the crop from the image
                current_image = current_image[rect[1] : rect[3], rect[0] : rect[2], :]

                # keep overlap with gt box IF center in sampled patch
                centers = (boxes[:, :2] + boxes[:, 2:]) / 2.0

                # mask in all gt boxes that above and to the left of centers
                m1 = (rect[0] < centers[:, 0]) * (rect[1] < centers[:, 1])

                # mask in all gt boxes that under and to the right of centers
                m2 = (rect[2] > centers[:, 0]) * (rect[3] > centers[:, 1])

                # mask in that both m1 and m2 are true
                mask = m1 * m2

                # have any valid boxes? try again if not
                if not mask.any():
                    continue

                # take only matching gt boxes
                current_boxes = boxes[mask, :].copy()

                # take only matching gt labels
                current_labels = labels[mask]

                # should we use the box left and top corner or the crop's
                current_boxes[:, :2] = numpy.maximum(current_boxes[:, :2], rect[:2])
                # adjust to crop (by substracting crop's left,top)
                current_boxes[:, :2] -= rect[:2]

                current_boxes[:, 2:] = numpy.minimum(current_boxes[:, 2:], rect[2:])
                # adjust to crop (by substracting crop's left,top)
                current_boxes[:, 2:] -= rect[:2]

                return current_image, current_boxes, current_labels


class CV2Expand(object):
    """ """

    def __init__(self, mean: float):
        self.mean = mean

    def __call__(
        self, image: numpy.ndarray, boxes: numpy.ndarray, labels: numpy.ndarray
    ) -> Tuple:
        if random.randint(2):
            return image, boxes, labels

        height, width, depth = image.shape
        ratio = random.uniform(1, 4)
        left = random.uniform(0, width * ratio - width)
        top = random.uniform(0, height * ratio - height)

        expand_image = numpy.zeros(
            (int(height * ratio), int(width * ratio), depth), dtype=image.dtype
        )
        expand_image[..., :] = self.mean
        expand_image[
            int(top) : int(top + height), int(left) : int(left + width)
        ] = image
        image = expand_image

        boxes = boxes.copy()
        boxes[:, :2] += (int(left), int(top))
        boxes[:, 2:] += (int(left), int(top))

        return image, boxes, labels


class CV2RandomMirror(object):
    def __call__(
        self, image: numpy.ndarray, boxes: numpy.ndarray, classes: numpy.ndarray
    ) -> Tuple:
        _, width, _ = image.shape
        if random.randint(2):
            image = image[:, ::-1]
            boxes = boxes.copy()
            boxes[:, 0::2] = width - boxes[:, 2::-2]
        return image, boxes, classes


class CV2SwapChannels(object):
    """Transforms a tensorized image by swapping the channels in the order
    specified in the swap tuple.
    Args:
    swaps (int triple): final order of channels
    eg: (2, 1, 0)"""

    def __init__(self, swaps: TripleNumber):
        self.swaps = swaps

    def __call__(self, image: numpy.ndarray) -> numpy.ndarray:
        """
        Args:
        image (Tensor): image tensor to be transformed
        Return:
        a tensor with channels swapped according to swap"""
        # if torch.is_tensor(image):
        #     image = image.data.cpu().numpy()
        # else:
        #     image = numpy.array(image)
        image = image[..., self.swaps]
        return image


class CV2PhotometricDistort(object):
    """ """

    def __init__(self):
        self.pd = [
            CV2RandomContrast(),  # RGB
            CV2ConvertColor(current="RGB", transform="HSV"),  # HSV
            CV2RandomSaturation(),  # HSV
            CV2RandomHue(),  # HSV
            CV2ConvertColor(current="HSV", transform="RGB"),  # RGB
            CV2RandomContrast(),  # RGB
        ]
        self.rand_brightness = CV2RandomBrightness()
        self.rand_light_noise = CV2RandomLightingNoise()

    def __call__(
        self, image: numpy.ndarray, boxes: numpy.ndarray, labels: numpy.ndarray
    ) -> Tuple:
        im = image.copy()
        im, boxes, labels = self.rand_brightness(im, boxes, labels)
        if random.randint(2):
            distort = CV2Compose(self.pd[:-1])
        else:
            distort = CV2Compose(self.pd[1:])
        im, boxes, labels = distort(im, boxes, labels)
        return self.rand_light_noise(im, boxes, labels)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 05-03-2021
           """

__all__ = ["increasing_density_hatch"]

from itertools import count

increasing_density_hatch = (b * a for a in count(1) for b in ("\\", "/", "x", "."))

four_times_denser_hatch = tuple(
    b * a for a in range(1, 4 + 1) for b in ("\\", "/", "x", ".")
)

if __name__ == "__main__":
    print(tuple(four_times_denser_hatch))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20-04-2021
           """

import itertools
from typing import Sequence

import numpy
from matplotlib import pyplot
from sklearn.metrics import classification_report, confusion_matrix
from sklearn.model_selection import train_test_split

__all__ = ["pretty_print_conf_matrix"]


def pretty_print_conf_matrix(
    y_true: Sequence,
    y_pred: Sequence,
    classes: Sequence,
    normalize: bool = False,
    title: str = "Confusion matrix",
    cmap=pyplot.cm.Blues,
) -> None:
    """
    Mostly stolen from: http://scikit-learn.org/stable/auto_examples/model_selection/plot_confusion_matrix.html#sphx-glr-auto-examples-model-selection-plot-confusion-matrix-py

    Normalization changed, classification_report stats added below plot
    """

    cm = confusion_matrix(y_true, y_pred)

    # Configure Confusion Matrix Plot Aesthetics (no text yet)
    pyplot.imshow(cm, interpolation="nearest", cmap=cmap)
    pyplot.title(title, fontsize=14)
    tick_marks = numpy.arange(len(classes))
    pyplot.xticks(tick_marks, classes, rotation=45)
    pyplot.yticks(tick_marks, classes)
    pyplot.ylabel("True label", fontsize=12)
    pyplot.xlabel("Predicted label", fontsize=12)

    # Calculate normalized values (so all cells sum to 1) if desired
    if normalize:
        cm = numpy.round(cm.astype("float") / cm.sum(), 2)  # (axis=1)[:, numpy.newaxis]

    # Place Numbers as Text on Confusion Matrix Plot
    thresh = cm.max() / 2.0
    for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
        pyplot.text(
            j,
            i,
            cm[i, j],
            horizontalalignment="center",
            color="white" if cm[i, j] > thresh else "black",
            fontsize=12,
        )

    # Add Precision, Recall, F-1 Score as Captions Below Plot
    rpt = classification_report(y_true, y_pred)
    rpt = rpt.replace("avg / total", "      avg")
    rpt = rpt.replace("support", "N Obs")

    pyplot.annotate(
        rpt,
        xy=(0, 0),
        xytext=(-50, -140),
        xycoords="axes fraction",
        textcoords="offset points",
        fontsize=12,
        ha="left",
    )

    # Plot
    pyplot.tight_layout()


if __name__ == "__main__":
    from sklearn import datasets
    from sklearn.svm import SVC

    # get data, make predictions
    (X, y) = datasets.load_iris(return_X_y=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, train_size=0.5)

    clf = SVC()
    clf.fit(X_train, y_train)
    y_test_pred = clf.predict(X_test)

    # Plot Confusion Matrix
    pyplot.style.use("classic")
    pyplot.figure(figsize=(3, 3))
    pretty_print_conf_matrix(
        y_test,
        y_test_pred,
        classes=["0", "1", "2"],
        normalize=True,
        title="Confusion Matrix",
    )
    pyplot.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

__doc__ = r"""
"""

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

try:
    from .spectral import *
    from .mpldrawer import *
    from .discrete_scroll_plot import *
    from .series_scroll_plot import *
    from .distribution_plot import *
    from .image_stream_plot import *
except ImportError as ix:
    print(f"Make sure requirements is installed for {Path(__file__).parent.name}")
    raise ix



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 29/07/2020
           """

__all__ = ["cross_validation_generator"]

from typing import Tuple

import torch
from sklearn.model_selection import KFold
from torch.utils.data import ConcatDataset, Dataset, Subset, TensorDataset

from draugr.torch_utilities.tensors.to_tensor import to_tensor


def cross_validation_generator(
    *datasets: Dataset, n_splits: int = 10
) -> Tuple[Subset, Subset]:
    """
    Learning the parameters of a prediction function and testing it on the same data is a methodological mistake: a model that would just repeat the labels of the samples that it has just seen would have a perfect score but would fail to predict anything useful on yet-unseen data. This situation is called overfitting. To avoid it, it is common practice when performing a (supervised) machine learning experiment to hold out part of the available data as a test set"""

    cum = ConcatDataset(datasets)
    for train_index, val_index in KFold(n_splits=n_splits).split(cum):
        yield Subset(cum, train_index), Subset(cum, val_index)


if __name__ == "__main__":

    def asdasidoj() -> None:
        """
        :rtype: None
        """
        X = to_tensor([torch.diag(torch.arange(i, i + 2)) for i in range(200)])
        x_train = TensorDataset(X[:100])
        x_val = TensorDataset(X[100:])

        for train, val in cross_validation_generator(x_train, x_val):
            print(len(train), len(val))
            print(train[0], val[0])

    asdasidoj()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/03/2020
           """

import getpass
import sys
from enum import Enum
from pathlib import Path

import sh  # pip install sh
from sorcery import assigned_names

from draugr import PROJECT_NAME
from draugr.os_utilities.linux_utilities.user_utilities import make_user, remove_user

__all__ = [
    "install_service",
    "remove_service",
    "enable_service",
    "disable_service",
    "start_service",
    "stop_service",
    "restart_service",
    "status_service",
    "RunAsEnum",
]

from draugr.os_utilities.linux_utilities.systemd_utilities.service_template import (
    SERVICE_TEMPLATE,
)
from warg import ContextWrapper


class RunAsEnum(Enum):
    """ """

    user, app_user, root = assigned_names()


def target_service_path(service_name, run_as: RunAsEnum = RunAsEnum.user):
    """ """
    if run_as == RunAsEnum.user:
        return Path.home() / ".config" / "systemd" / "user" / f"{service_name}.service"
    elif run_as == run_as.root:
        return Path("/") / "etc" / "systemd" / "system" / f"{service_name}.service"
    return Path("/") / "lib" / "systemd" / "system" / f"{service_name}.service"


class RestartServiceEnum(Enum):
    """
    Restart settings\Exit causes	 no	  always      on-success	    on-failure	    on-abnormal	on-abort	on-watchdog
    Clean exit code or signal	 	        X	          X
    Unclean exit code	 	                X	 	                        X
    Unclean signal	 	                  X	 	                        X	              X	          X
    Timeout	 	                          X	 	                        X	              X
    Watchdog	 	                        X	 	                        X	              X	 	                  X
    """

    no, always = assigned_names()
    on_failure = "on-failure"
    on_success = "on-success"
    on_abnormal = "on-abnormal"
    on_abort = "on-abort"
    on_watchdog = "on-watchdog"


class ServiceTargetEnum(Enum):  # TODO: UTILISE!
    """
    Default is default

    SysV 	      systemd Target	                                                Notes
    0	          runlevel0.target, poweroff.target	                              Halt the system.
    1, s,       single	runlevel1.target, rescue.target	                        Single user mode.
    2, 4	      runlevel2.target, runlevel4.target, multi-user.target	          User-defined/Site-specific runlevels. By default, identical to 3.
    3	          runlevel3.target, multi-user.target	                            Multi-user, non-graphical. Users can usually login via multiple consoles or via the network.
    5	          runlevel5.target, graphical.target	                            Multi-user, graphical. Usually has all the services of runlevel 3 plus a graphical login.
    6	          runlevel6.target, reboot.target	                                Reboot
    emergency	  emergency.target	                                              Emergency shell
    """

    default = "default.target"
    multi_user = "multi-user.target"


def install_service(
    service_entry_point_path: Path,
    service_name: str,
    *,
    description: str = None,
    auto_enable: bool = True,
    run_as: RunAsEnum = RunAsEnum.user,
    # get_sudo: bool = False,
    restart: RestartServiceEnum = RestartServiceEnum.on_failure,
) -> None:
    """
    Args:
        :param restart:
        :param service_entry_point_path:
        :param service_name:
        :param description:
        :param auto_enable:
        :param run_as:
    """
    assert (
        service_entry_point_path.is_file()
        and service_entry_point_path.name.endswith(".py")
    )
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    user = getpass.getuser()

    systemd_service_file_path = target_service_path(project_service_name, run_as=run_as)
    print(f"Installing {systemd_service_file_path}")
    get_sudo = run_as != RunAsEnum.user
    with ContextWrapper(
        sh.contrib.sudo,
        construction_kwargs=dict(
            password=(
                getpass.getpass(prompt=f"[sudo] password for {user}: ")
                if get_sudo
                else None
            ),
            _with=True,
        ),
        enabled=get_sudo,
    ):
        if run_as == RunAsEnum.app_user:
            service_user = service_name + "_user"
            make_user(service_user, get_sudo=False)
            service_target = "default.target"
            service_group = service_user
        elif run_as == RunAsEnum.root:
            service_user = "root"
            service_target = "multi-user.target"
            service_group = service_user
        elif run_as == RunAsEnum.user:
            service_user = user
            service_target = "default.target"
            service_group = service_user
        else:
            raise ValueError

        sh.touch(systemd_service_file_path)
        group_name = str(sh.id(["-g", "-n", service_user])).strip("\n")
        assert service_group == group_name
        current_owner = sh.ls("-l", systemd_service_file_path).split(" ")[2]
        if current_owner != service_user:  # SETTING UP PERMISSIONS
            print(
                f"Changing owner of service file from {current_owner} to {service_user}"
            )
            if run_as == RunAsEnum.root:
                group_name = ""
            else:
                print(f"with common group {group_name}")
                # group_id = sh.id(["-g", service_user])
                sh.usermod(
                    ["-a", "-G", group_name, user]
                )  # TODO: Polluting groups of user
        sh.chown(
            [f"{user}:{group_name}", service_entry_point_path]
        )  # If a colon but no group name follows the user name, that user is made the owner of the files and the group of the files is changed to that user's login group.
        sh.chown(
            [f"{user}:{group_name}", systemd_service_file_path]
        )  # If a colon but no group name follows the user name, that user is made the owner of the files and the group of the files is changed to that user's login group.

        print("writing service file")
        if not description:
            description = f"heimdallr service for {service_name}"
        with open(systemd_service_file_path, "w") as f:
            f.writelines(
                SERVICE_TEMPLATE.format(
                    service_name=project_service_name,
                    service_user=service_user,
                    executable=sys.executable,
                    description=description,
                    service_entry_point_path=service_entry_point_path,
                    service_target=service_target,
                    service_group=service_group,
                    restart=restart.value,
                )
            )
        sh.chown(
            [f"{service_user}:{group_name}", systemd_service_file_path]
        )  # If a colon but no group name follows the user name, that user is made the owner of the files and the group of the files is changed to that user's login group.
        sh.chmod(["664", systemd_service_file_path])
        sh.chmod(["774", service_entry_point_path])
        sh.systemctl("daemon-reload")  # TODO: Requires sudo?

        if auto_enable:
            enable_service(service_name, get_sudo=False, run_as=run_as)


def remove_service(
    service_name: str,
    *,
    remove_app_user: bool = True,
    get_sudo: bool = False,
    run_as: RunAsEnum = RunAsEnum.user,
) -> None:
    """

    Args:
        :param get_sudo:
        :param service_name:
        :param remove_app_user:
        :param run_as:
    """

    try:
        # get_sudo = not run_as == RunAsEnum.user
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=(
                    getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                    if get_sudo
                    else None
                ),
                _with=True,
            ),
            enabled=get_sudo,
        ):
            disable_service(service_name, get_sudo=False, run_as=run_as)
            project_service_name = f"{PROJECT_NAME}_service_{service_name}"
            target_service_file_path = target_service_path(
                project_service_name, run_as=run_as
            )
            print(f"Removing {target_service_file_path}")

            sh.rm(target_service_file_path)
            sh.systemctl("daemon-reload")

            if run_as == RunAsEnum.app_user and remove_app_user:
                # DO CLEAN UP!
                remove_user(service_name + "_user", get_sudo=False, run_as=run_as)
    except sh.ErrorReturnCode_1:
        pass


def enable_service(
    service_name: str, *, get_sudo: bool = False, run_as: RunAsEnum = RunAsEnum.user
) -> None:
    """

    Args:
        service_name:
        :param service_name:
        :param run_as:
        :param get_sudo:
    """
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    print(f"Enabling {project_service_name}")
    with ContextWrapper(
        sh.contrib.sudo,
        construction_kwargs=dict(
            password=(
                getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                if get_sudo
                else None
            ),
            _with=True,
        ),
        enabled=get_sudo,
    ):
        sh.systemctl(
            (["--user"] if run_as == RunAsEnum.user else [])
            + [f"enable", f"{project_service_name}.service"]
        )
        start_service(service_name, get_sudo=False, run_as=run_as)


def disable_service(
    service_name: str, *, get_sudo: bool = False, run_as: RunAsEnum = RunAsEnum.user
) -> None:
    """

    Args:
        service_name:
        :param service_name:
        :param run_as:
        :param get_sudo:
    """
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    print(f"Disabling {project_service_name}")
    try:
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=(
                    getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                    if get_sudo
                    else None
                ),
                _with=True,
            ),
            enabled=get_sudo,
        ):
            stop_service(service_name, get_sudo=False, run_as=run_as)
            sh.systemctl(
                (["--user"] if run_as == RunAsEnum.user else [])
                + ["disable", f"{project_service_name}.service"]
            )
    except sh.ErrorReturnCode_5:
        pass


def stop_service(
    service_name: str, *, get_sudo: bool = False, run_as: RunAsEnum = RunAsEnum.user
) -> None:
    """

    Args:
        service_name:
        :param service_name:
        :param run_as:
        :param get_sudo:
    """
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    print(f"Stopping {project_service_name}")
    try:
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=(
                    getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                    if get_sudo
                    else None
                ),
                _with=True,
            ),
            enabled=get_sudo,
        ):
            sh.systemctl(
                (["--user"] if run_as == RunAsEnum.user else [])
                + ["stop", f"{project_service_name}.service"]
            )
    except sh.ErrorReturnCode_5:
        pass


def start_service(
    service_name: str, *, get_sudo: bool = False, run_as: RunAsEnum = RunAsEnum.user
) -> None:
    """

    Args:
        service_name:
        :param service_name:
        :param run_as:
        :param get_sudo:
    """
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    print(f"Starting {project_service_name}")
    try:
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=(
                    getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                    if get_sudo
                    else None
                ),
                _with=True,
            ),
            enabled=get_sudo,
        ):
            sh.systemctl(
                (["--user"] if run_as == RunAsEnum.user else [])
                + ["start", f"{project_service_name}.service"]
            )
    except sh.ErrorReturnCode_5:
        pass


def restart_service(
    service_name: str, *, get_sudo: bool = False, run_as: RunAsEnum = RunAsEnum.user
) -> None:
    """

    Args:
        service_name:
        :param service_name:
        :param run_as:
        :param get_sudo:
    """
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    print(f"Restarting {project_service_name}")
    try:
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=(
                    getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                    if get_sudo
                    else None
                ),
                _with=True,
            ),
            enabled=get_sudo,
        ):
            sh.systemctl(
                (["--user"] if run_as == RunAsEnum.user else [])
                + ["restart", f"{project_service_name}.service"]
            )
    except sh.ErrorReturnCode_5:
        pass


def status_service(
    service_name: str, *, get_sudo: bool = False, run_as: RunAsEnum = RunAsEnum.user
) -> None:
    """

    Args:
        service_name:
        :param service_name:
        :param run_as:
        :param get_sudo:
    """
    project_service_name = f"{PROJECT_NAME}_service_{service_name}"
    print(f"Status for {project_service_name}")
    try:
        with ContextWrapper(
            sh.contrib.sudo,
            construction_kwargs=dict(
                password=(
                    getpass.getpass(prompt=f"[sudo] password for {getpass.getuser()}: ")
                    if get_sudo
                    else None
                ),
                _with=True,
            ),
            enabled=get_sudo,
        ):
            sh.systemctl(
                (["--user"] if run_as == RunAsEnum.user else [])
                + ["status", f"{project_service_name}.service"]
            )
    except sh.ErrorReturnCode_3 as e:
        print(e, e.stdout)


if __name__ == "__main__":
    pass
    print(RunAsEnum.user.value)
    # remove_service("busy_script")
    # install_service('busy_script')
    # status_service('busy_script')
    # restart_service('busy_script')



import argparse
from pathlib import Path
from typing import Any, List, Tuple

import numpy
from PIL import Image

from draugr.python_utilities.colors import color


def get_pixel(col: Tuple) -> Any:
    """ """
    if isinstance(col, (float, int)) or len(col) == 1:
        col = (col, col, col)
    return color("  ", bg=f"rgb({int(col[0])}, {int(col[1])}, {int(col[2])})")


def render_image(pixels: numpy.ndarray, scale: Tuple) -> List[List]:
    """ """
    # first of all scale the image to the scale 'tuple'
    image_size = pixels.shape[:2]
    block_size = (image_size[0] / scale[0], image_size[1] / scale[1])

    blocks = []
    y = 0
    while y < image_size[0]:
        x = 0
        block_col = []
        while x < image_size[1]:
            # get a block, reshape in into an Nx3 matrix and then get average of each column
            block_col.append(
                pixels[int(y) : int(y + block_size[0]), int(x) : int(x + block_size[1])]
                .reshape(-1, 3)
                .mean(axis=0)
            )
            x += block_size[1]
        blocks.append(block_col)
        y += block_size[0]
    output = [[get_pixel(block) for block in row] for row in blocks]
    return output


def terminalise_image(output):
    """

    joins lists

    :return:"""
    return "\n".join(["".join(row) for row in output])


def get_image(path: Path):
    """ """
    img = numpy.asarray(Image.open(path))
    if img.shape[2] > 3:
        return numpy.array([[pixel[:3] for pixel in row] for row in img])
    return img


def render_file(path: Path, scale=(60, 60)):
    """ """
    image = get_image(path)
    output = render_image(image, scale)
    print(terminalise_image(output))


def entry_point() -> None:
    """
    :rtype: None
    """
    parser = argparse.ArgumentParser(description="Render images on the command line")
    parser.add_argument("path", metavar="path", type=str, help="the image path")
    parser.add_argument(
        "--width",
        dest="width",
        default=60,
        type=int,
        help="width of the rendered image (default 60 pixels)",
    )
    parser.add_argument(
        "--height",
        dest="height",
        default=60,
        type=int,
        help="height of the rendered image (default 60 pixels)",
    )
    args = parser.parse_args()
    render_file(args.path, (args.height, args.width))


if __name__ == "__main__":
    render_file(Path.home() / "OneDrive" / "Billeder" / "pompey.jpg", scale=(20, 20))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-05-2021
           """

from draugr.os_utilities.resource_utilities.cpu import worker_cores_available
from draugr.os_utilities.resource_utilities.ram import num_instance_no_paging


def get_num_instances(expected_size: int = 1024) -> int:
    """

    :param expected_size:
    :return:
    """
    return min(worker_cores_available(), num_instance_no_paging(expected_size))


if __name__ == "__main__":
    print(get_num_instances())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/01/2020
           """

from .concatination import *
from .disjunction import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 06-04-2021
           """

from pathlib import Path

import torch
from torch.optim.optimizer import Optimizer

__all__ = ["save_checkpoint", "load_checkpoint"]


def save_optimiser(
    *,
    optimiser: Optimizer,
    optimiser_save_path: Path,
    raise_on_existing: bool = False,
) -> None:
    """

    :param optimiser:
    :param optimiser_save_path:
    :param raise_on_existing:"""
    if raise_on_existing and optimiser_save_path.exists():
        raise FileExistsError(f"{optimiser_save_path} exists!")
    torch.save(optimiser, str(optimiser_save_path))


def save_checkpoint(PATH: Path, epoch, model, optimiser, loss):
    """

    :param PATH:
    :param epoch:
    :param model:
    :param optimiser:
    :param loss:
    """
    torch.save(
        {
            "epoch": epoch,
            "model_state_dict": model.state_dict(),
            "optimiser_state_dict": optimiser.state_dict(),
            "loss": loss,
        },
        PATH,
    )

    PATH.with_suffix(".tar")


def load_checkpoint(PATH: Path, model, optimizer):
    """

    :param PATH:
    :param model:
    :param optimizer:
    :return:
    """
    checkpoint = torch.load(PATH)

    epoch = checkpoint["epoch"]
    model.load_state_dict(checkpoint["model_state_dict"])
    optimizer.load_state_dict(checkpoint["optimiser_state_dict"])
    loss = checkpoint["loss"]

    return epoch, model, optimizer, loss


if __name__ == "__main__":

    def main() -> None:
        """
        :rtype: None
        """
        pass
        # model = TheModelClass(args, **kwargs)
        # optimizer = TheOptimizerClass(args, **kwargs)

    def multi() -> None:
        """
        :rtype: None
        """
        pass

    # checkpoint = torch.load(PATH)
    # modelA.load_state_dict(checkpoint['modelA_state_dict'])
    # modelB.load_state_dict(checkpoint['modelB_state_dict'])
    # optimizerA.load_state_dict(checkpoint['optimizerA_state_dict'])
    # optimizerB.load_state_dict(checkpoint['optimizerB_state_dict'])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03-05-2021
           """

from .dlib_utilities import *
from .facealigner import *



from enum import Enum

import cv2
from sorcery import assigned_names

from draugr.opencv_utilities.namespaces.enums import MorphShapeEnum, MorphTypeEnum
from warg import next_odd

__all__ = ["to_edge", "ToEdgeMethodEnum", "CannyApertureSize"]


class ToEdgeMethodEnum(Enum):
    canny, laplacian, sobel_vh, sobel_h, sobel_v, morph_gradient = assigned_names()


class CannyApertureSize(Enum):
    a3, a5, a7 = 3, 5, 7


def to_edge(img, method: ToEdgeMethodEnum = ToEdgeMethodEnum.canny, **kwargs):
    method = ToEdgeMethodEnum(method)
    if method == ToEdgeMethodEnum.canny:
        return cv2.Canny(
            img,
            threshold1=kwargs.get("threshold1", 60),
            threshold2=kwargs.get("threshold2", 180),
            apertureSize=kwargs.get("apertureSize", CannyApertureSize.a3).value,
            L2gradient=kwargs.get("L2gradient", None),
        )

    ksize = kwargs.get("ksize", max(next_odd(max(*(img.shape)) // 100), 5))
    if method == ToEdgeMethodEnum.morph_gradient:

        return cv2.morphologyEx(
            img,
            MorphTypeEnum.gradient.value,
            cv2.getStructuringElement(MorphShapeEnum.rect.value, ksize=(ksize, ksize)),
        )
    elif method == ToEdgeMethodEnum.laplacian:
        return cv2.Laplacian(
            img, cv2.CV_8UC1, ksize=ksize
        )  # ,cv2.CV_16UC1, #cv2.CV_16S, # cv2.CV_64F
    elif method == ToEdgeMethodEnum.sobel_h:
        return cv2.Sobel(img, cv2.CV_64F, 1, 0, ksize=ksize)
    elif method == ToEdgeMethodEnum.sobel_v:
        return cv2.Sobel(img, cv2.CV_64F, 0, 1, ksize=ksize)
    elif method == ToEdgeMethodEnum.sobel_vh:
        sobelx = cv2.Sobel(img, cv2.CV_64F, 1, 0, ksize=ksize)
        sobely = cv2.Sobel(img, cv2.CV_64F, 0, 1, ksize=ksize)
        return sobelx + sobely

    raise NotImplementedError


if __name__ == "__main__":

    def aushd():
        import numpy
        from draugr.opencv_utilities import show_image

        a = numpy.zeros((50, 50))
        a[:, 25] = 1
        a[25, :] = 1

        show_image(to_edge(a, ToEdgeMethodEnum.sobel_vh), wait=True)

    aushd()



# !/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 06-01-2021
           """

from typing import Sequence

import mpl_toolkits.mplot3d.axes3d as p3
import numpy
from matplotlib import animation, cm, pyplot
from mpl_toolkits.mplot3d import axes3d
from scipy.signal import chirp, spectrogram

from warg import next_pow_2

__all__ = ["spectral_plot3d", "spectrum_plot3d"]


# TODO: ANIMATED VARIANT, maybe as a drawer!


def spectral_plot3d(
    time: numpy.ndarray, frequencies: numpy.ndarray, fxt: numpy.ndarray
) -> pyplot.Figure:
    """
    return new figure"""
    assert fxt.shape == (*frequencies.shape, *time.shape)
    assert fxt.dtype == numpy.complex

    fig = pyplot.figure()
    ax = p3.Axes3D(fig)

    x, y = numpy.meshgrid(time, frequencies)

    # colors=cm.jet(norm(colorfunction))
    colors = numpy.empty(x.shape, dtype=numpy.float)
    z = numpy.empty(x.shape, dtype=numpy.float)

    for y_i in range(len(time)):
        for x_i in range(len(frequencies)):
            com = fxt[x_i, y_i]
            z[x_i, y_i] = com.real
            colors[x_i, y_i] = com.imag * 0.5 + 0.5

    colors = colors / colors.max()

    surf = ax.plot_surface(
        x,
        y,
        z,
        facecolors=cm.jet(colors),
        # linewidth=0
        # color='0.75',
        # rstride=1,
        # cstride=1
        # rcount=50
        # #ccount=50
    )

    ax.set_ylabel("Frequency [kHz]")
    ax.set_xlabel("Time [s]")
    ax.set_zlabel("Magnitude")
    return fig


def spectrum_plot3d(
    signal: Sequence, sampling_rate: int, window_length_ms=(20 / 1000)
) -> pyplot.Figure:
    """ """
    n_per_seg = next_pow_2(
        sampling_rate * window_length_ms
    )  # 20 ms, next_pow_2 per seg == n_fft
    f, t, fxt = spectrogram(
        signal,
        fs=sampling_rate,
        window="hanning",
        nperseg=n_per_seg,
        scaling="spectrum",
        mode="complex",
    )

    return spectral_plot3d(t, f, fxt)


if __name__ == "__main__":

    def asdijaisd() -> None:
        """
        :rtype: None
        """
        sr = 1000
        t = numpy.arange(sr * 4) / sr
        # noise = numpy.random.rand(sr * 2) * 0.001
        w = chirp(t, f0=100, f1=500, t1=4, method="linear")
        signal = numpy.sin(200 * 2 * numpy.pi * t) + w  # + noise

        spectrum_plot3d(signal, sr)
        pyplot.show()

    def aisjd() -> None:
        """
        :rtype: None
        """
        fig = pyplot.figure()
        ax = axes3d.Axes3D(fig)

        def gen(n):
            """ """
            phi = 0
            while phi < 2 * numpy.pi:
                yield numpy.array([numpy.cos(phi), numpy.sin(phi), phi])
                phi += 2 * numpy.pi / n

        def update(num, data, line):
            """ """
            line.set_data(data[:2, :num])
            line.set_3d_properties(data[2, :num])

        N = 100
        data = numpy.array(list(gen(N))).T
        (line,) = ax.plot(data[0, 0:1], data[1, 0:1], data[2, 0:1])

        # Setting the axes properties
        ax.set_xlim3d([-1.0, 1.0])
        ax.set_xlabel("X")

        ax.set_ylim3d([-1.0, 1.0])
        ax.set_ylabel("Y")

        ax.set_zlim3d([0.0, 10.0])
        ax.set_zlabel("Z")

        ani = animation.FuncAnimation(
            fig, update, N, fargs=(data, line), interval=10000 / N, blit=False
        )
        # ani.save('matplot003.gif', writer='imagemagick')
        pyplot.show()

    # aisjd()
    asdijaisd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-03-2021
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .event_export import *
from .file_export import *
from .db_export import *

if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"
__doc__ = r"""
"""

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .truncation import *
from .signal_statistics import *
from .signal_mixing import *
from .signal_projection import *
from .padding import *



import warnings

import cv2

from draugr.opencv_utilities.namespaces.color_conversion_enum import (
    ColorConversionEnum,
)

__all__ = ["num_channels", "is_singular_channel", "to_color"]


def num_channels(image) -> int:
    return image.shape[-1] if image.ndim == 3 else 1


def is_singular_channel(image) -> bool:
    return num_channels(image) == 1


def to_color(image, conversion: ColorConversionEnum = ColorConversionEnum.gray2bgr):
    """
    convert from the default bgr cv2 format to gray, using a single component

    :param image:
    :type image:
    :param component:
    :type component:
    :param to_gray_method:
    :type to_gray_method:
    :return:
    :rtype:
    """

    if is_singular_channel(image):
        conversion = ColorConversionEnum(conversion)
        image = cv2.cvtColor(image, conversion)
    else:
        warnings.warn(f"expected singular image was shape {image.shape}")
    return image



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Any, Iterable, List, Sequence

import numpy
import torch
from draugr.torch_utilities.tensors.to_tensor import to_tensor
from draugr.torch_utilities.architectures.mlp import MLP

__author__ = "Christian Heider Nielsen"
__doc__ = "Fusion variant of MLPs"

__all__ = ["PreConcatInputMLP", "LateConcatInputMLP"]

from warg import passes_kws_to


class PreConcatInputMLP(MLP):
    """
    Early fusion
    """

    def __init__(self, input_shape: Sequence = (2,), **kwargs):
        if isinstance(input_shape, Iterable):
            input_shape = sum(input_shape)

        super().__init__(input_shape=input_shape, **kwargs)

    @passes_kws_to(MLP.forward)
    def forward(self, *x, **kwargs) -> List:
        """

        :param x:
        :type x:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        return super().forward(torch.cat(x, dim=-1), **kwargs)


class LateConcatInputMLP(MLP):
    """
    Late fusion, quite a botch job, only a single addition block fusion supported for now
    You have been warned! ;)
    """

    def __init__(
        self, input_shape: Sequence = (2, 2), output_shape: Sequence = (2,), **kwargs
    ):

        forward_shape, *res = input_shape
        self._residual_shape = res

        if not isinstance(self._residual_shape, Iterable):
            self._residual_shape = (self._residual_shape,)

        if not isinstance(output_shape, Iterable):
            output_shape = (*self._residual_shape, output_shape)
        assert len(output_shape) == 2

        super().__init__(
            input_shape=(forward_shape,), output_shape=output_shape, **kwargs
        )

        s = sum((*output_shape, *self._residual_shape))
        t = s * 10  # Hidden
        self.post_concat_layer = torch.nn.Sequential(
            torch.nn.Linear(s, t), torch.nn.ReLU(), torch.nn.Linear(t, output_shape[-1])
        )

    @passes_kws_to(MLP.forward)
    def forward(self, *x, **kwargs) -> torch.tensor:
        """

        :param x:
        :type x:
        :param kwargs:
        :type kwargs:
        :return:
        :rtype:
        """
        forward_x, *residual_x = x
        return self.post_concat_layer(
            torch.cat((*super().forward(forward_x, **kwargs), *residual_x), dim=-1)
        )


if __name__ == "__main__":

    def stest_normal():
        """ """
        s = (10,)
        a = (10,)
        model = PreConcatInputMLP(input_shape=s, output_shape=a)

        inp = to_tensor(range(s[0]), device="cpu")
        print(model.forward(inp))

    def stest_multi_dim_normal():
        """ """
        s = (19,)
        s1 = (4,)
        batch_size = (100,)
        a = (2, 10)
        model = PreConcatInputMLP(input_shape=s + s1, output_shape=a)

        inp = to_tensor(numpy.random.random((*batch_size, *s)), device="cpu")
        late_input = to_tensor(numpy.random.random((*batch_size, *s1)), device="cpu")
        print(model.forward(inp, late_input))

    def stest_multi_dim_normal21():
        """ """
        s = (19,)
        s1 = (4,)
        batch_size = (100,)
        a = (2, 10)
        model = LateConcatInputMLP(input_shape=s + s1, output_shape=a)

        inp = to_tensor(numpy.random.random((*batch_size, *s)), device="cpu")
        late_input = to_tensor(numpy.random.random((*batch_size, *s1)), device="cpu")
        print(model.forward(inp, late_input))

    def stest_multi_dim_normal23121():
        """ """
        s = (19,)
        s1 = (4,)
        batch_size = (100,)
        output_shape = (1, 2)
        model = LateConcatInputMLP(input_shape=s + s1, output_shape=output_shape)

        inp = to_tensor(numpy.random.random((*batch_size, *s)), device="cpu")
        late_input = to_tensor(numpy.random.random((*batch_size, *s1)), device="cpu")
        print(model.forward(inp, late_input))

    def stest_multi_dim_normal2321412121():
        """ """
        s = (19,)
        s1 = (4,)
        batch_size = (100,)
        output_shape = 2
        model = LateConcatInputMLP(input_shape=s + s1, output_shape=output_shape)

        inp = to_tensor(numpy.random.random((*batch_size, *s)), device="cpu")
        late_input = to_tensor(numpy.random.random((*batch_size, *s1)), device="cpu")
        print(model.forward(inp, late_input))

    stest_normal()
    stest_multi_dim_normal()
    stest_multi_dim_normal21()
    stest_multi_dim_normal23121()
    stest_multi_dim_normal2321412121()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11-12-2020
           """

from pathlib import Path

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .shallow_category import *
from .deep_category import *
from .flat_category import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03/04/2020
           """

from .fast_fourier_transform import *
from .fast_fourier_transform_spectrogram import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

import numpy

__all__ = [
    "gamma_correct_byte",
    "gamma_correct_fast_to_byte",
    "gamma_correct_float_to_byte",
    "linear_correct_byte",
    "linear_correct_float_to_byte",
]


# from numba import njit # PERFORM EXPERIMENTS


# @njit
def gamma_correct_byte(image: numpy.ndarray, gamma: float = 2.2) -> numpy.ndarray:
    """

    :param image:
    :type image:
    :param gamma:
    :type gamma:
    :return:
    :rtype:"""
    return gamma_correct_float_to_byte(image / 255, gamma)


# @njit
def gamma_correct_fast_to_byte(image: numpy.ndarray) -> numpy.ndarray:
    """

    :param image:
    :type image:
    :return:
    :rtype:"""
    return ((image**0.454545) * 255).astype(numpy.uint8)


# @njit
def gamma_correct_float_to_byte(
    image: numpy.ndarray, gamma: float = 2.2
) -> numpy.ndarray:
    """

    :param image:
    :type image:
    :param gamma:
    :type gamma:
    :return:
    :rtype:"""
    return ((image ** (1.0 / gamma)) * 255).astype(numpy.uint8)


# @njit
def linear_correct_float_to_byte(
    image: numpy.ndarray, gamma: float = 2.2
) -> numpy.ndarray:
    """

    :param image:
    :type image:
    :param gamma:
    :type gamma:
    :return:
    :rtype:"""
    return ((image**gamma) * 255).astype(numpy.uint8)


# @njit
def linear_correct_byte(image: numpy.ndarray, gamma: float = 2.2) -> numpy.ndarray:
    """

    :param image:
    :type image:
    :param gamma:
    :type gamma:
    :return:
    :rtype:"""
    return gamma_correct_float_to_byte(image / 255, gamma)


if __name__ == "__main__":
    inp = numpy.random.random((2, 2))
    print(gamma_correct_fast_to_byte(inp))
    print(gamma_correct_float_to_byte(inp))
    print(linear_correct_float_to_byte(inp))
    print(linear_correct_byte(inp))
    print(gamma_correct_byte(inp))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Any, Sequence

import torch
from torch import nn

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
"""
__all__ = ["Architecture"]

from draugr.torch_utilities.optimisation.parameters import get_num_parameters
from warg import drop_unused_kws
from draugr import indent_lines


class Architecture(nn.Module, ABC):
    """ """

    @drop_unused_kws
    def __init__(self, input_shape: Sequence[int], output_shape: Sequence[int]):
        super().__init__()
        self._input_shape = input_shape
        self._output_shape = output_shape

    @property
    def input_shape(self) -> Sequence[int]:
        """

        :return:
        :rtype:"""
        return self._input_shape

    @property
    def output_shape(self) -> Sequence[int]:
        """

        :return:
        :rtype:"""
        return self._output_shape

    def sample_input(self) -> Any:
        return torch.empty(1, *self.input_shape, device="cpu")

    def __repr__(self):
        num_trainable_params = get_num_parameters(self, only_trainable=True)
        num_params = get_num_parameters(self, only_trainable=False)

        dict_repr = indent_lines(f"{self.__dict__}")

        trainable_params_str = indent_lines(
            f"trainable/num_params: {num_trainable_params}/{num_params}\n"
        )

        return f"{super().__repr__()}\n{dict_repr}\n{trainable_params_str}"


if __name__ == "__main__":
    a = Architecture()

    print(a)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-09-2020
           """

from typing import Any

import matlab
import numpy
from scipy import sparse

__all__ = ["ndarray_to_matlab", "matlab_to_ndarray", "dict_to_sparse"]

from draugr.matlab_utilities.matlab_conversion_utilities import (
    get_strides_c,
    get_strides_f,
)


def matlab_to_ndarray(matlab_array: Any) -> numpy.ndarray:
    """
    Convert matlab mlarray to numpy array
    The conversion is realised without copy for real data thank to the frombuffer
    protocol. The np ndarray type depends on the type of matlab data.

    Paramerters
    -----------
    matlab_array : mlarray
    the matlab array to convert

    Returns
    -------
    numpy_array : numpy array
    the converted numpy array

    Examples
    --------
    Complex 2D array
    >>> mc=matlab.double([[(1+1j),(0.3+0j),1j],[(-1+1.2j),0j,(1+1j)]], is_complex=True)
    >>> matlab_to_ndarray(mc)
    array([[ 1. +1.j ,  0.3+0.j ,  0. +1.j ],
       [-1. +1.2j,  0. +0.j ,  1. +1.j ]])
    >>> m3 = matlab.double([[[1.0,10.0],[2.0,20.0],[3.0,30.0]],[[4.0,40.0],[5.0,50.0],[6.0,60.0]]])
    >>> np3 = matlab_to_ndarray(m3)
    >>> np3[...,0]
    array([[1., 2., 3.],
       [4., 5., 6.]])
    >>> np3[...,1]
    array([[10., 20., 30.],
       [40., 50., 60.]])
    >>> np3.flags['OWNDATA'] # no copy
    False
    References
    ----------
    https://stackoverflow.com/questions/34155829/how-to-efficiently-convert-matlab-engine-arrays-to-numpy-ndarray/34155926


    for real:
    ma._data.itemsize # get item size
    ma._data.typecode # get item type
    ma._python_type
    https://docs.scipy.org/doc/numpy/reference/generated/numpy.dtype.html#numpy.dtype
    https://docs.python.org/3/library/array.html

    numpy kind

    A character code (one of ‘biufcmMOSUV’) identifying the general kind of data.
    b 	boolean
    i 	signed integer
    u 	unsigned integer
    f 	floating-point
    c 	complex floating-point
    m 	timedelta
    M 	datetime
    O 	object
    S 	(byte-)string
    U 	Unicode
    V 	void


    :param matlab_array:
    :return:"""

    if "mlarray" not in str(type(matlab_array)):
        raise TypeError(f"Expected matlab.mlarray. Got {type(matlab_array)}")

    if matlab_array._is_complex:
        numpy_type = "f8"
        numpy_array = numpy.empty(matlab_array.size, dtype=complex)
        numpy_array.real = numpy.frombuffer(
            matlab_array._real, dtype=numpy_type
        ).reshape(matlab_array.size, order="F")
        numpy_array.imag = numpy.frombuffer(
            matlab_array._imag, dtype=numpy_type
        ).reshape(matlab_array.size, order="F")
        return numpy_array

    # tuple that define type
    if isinstance(matlab_array._data, numpy.ndarray):
        return matlab_array._data.reshape(matlab_array.size)

    matlab_type = (
        matlab_array._data.typecode,
        matlab_array._data.itemsize,
    )  # use test to define the type, note not comprehensive a few types are missing, like uint!!
    if matlab_type == ("d", 8):  # double
        numpy_type = "f8"
    elif matlab_type == ("B", 1):  # logical is given as a int
        numpy_type = "bool"
    elif matlab_type == ("b", 1):  # int8
        numpy_type = "i1"
    elif matlab_type == ("i", 4):
        numpy_type = "i4"
    elif matlab_type == ("i", 8):  # int64
        numpy_type = "i8"
    elif matlab_type == ("l", 8):  # int64
        numpy_type = "i8"
    else:
        numpy_type = "f8"  # default

    return numpy.frombuffer(matlab_array._data, dtype=numpy_type).reshape(
        matlab_array.size, order="F"
    )  # no copy with the buffer


def ndarray_to_matlab(numpy_array):
    """Conversion of a numpy array to matlab mlarray

    The conversion is realised without copy for real data. First an empty initialization is realized.
    Then the numpy array is affected to the _data field. Thus the data field is not really an
    array.array but a numpy array. Matlab doesn't see anything...
    For complex data, the strides seems to not work properly with matlab.double.

    Paramerters
    -----------
    numpy_array : numpy array
      the array to convert
    Returns
    -------
    matlab_array : mlarray
      the converted array that can be passe to matlab
    Examples
    --------
    >>> npi=numpy.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12]],dtype=numpy.int64,order='C')
    >>> ndarray_to_matlab(npi)
    matlab.int64([[1,2,3],[4,5,6],[7,8,9],[10,11,12]])
    >>> npif=numpy.array([[1,2,3],[4,5,6],[7,8,9],[10,11,12]],dtype=numpy.int64,order='F')
    >>> ndarray_to_matlab(npif)
    matlab.int64([[1,2,3],[4,5,6],[7,8,9],[10,11,12]])

    >>> npcf=numpy.array([[1,2+0.2j,3],[4,5,6],[7,8,9],[10+0.1j,11,12]],dtype=numpy.complex,order='F')
    >>> ndarray_to_matlab(npcf)
    matlab.double([[(1+0j),(2+0.2j),(3+0j)],[(4+0j),(5+0j),(6+0j)],[(7+0j),(8+0j),(9+0j)],[(10+0.1j),(11+0j),(12+0j)]], is_complex=True)

    References
    -----------
    https://scipy-lectures.org/advanced/advanced_numpy/ (strides)
    """

    if "ndarray" not in str(type(numpy_array)):  # check numpy
        raise TypeError(f"Expect  numpy.ndarray. Got {type(numpy_array)}")

    shape = numpy_array.shape  # get shape

    num_elements = numpy.prod(shape)  # number of elements

    if numpy_array.flags.f_contiguous:  # compute strides (real case)
        strides = get_strides_f(shape)
        order = "F"
    else:
        strides = get_strides_c(shape)
        order = "C"

    if numpy.iscomplexobj(numpy_array):  # complex case

        matlab_array = matlab.double(
            initializer=None, size=(1, num_elements), is_complex=True
        )  # create empty matlab.mlarray
        # associate the data
        """
# associate the data (no copy), works on 2D array only...
matlab_array._real=numpy_array.ravel(order=order) # allow to map real and imaginary part continuously!
"""
        cpx = numpy_array.ravel(order="F")  # copy except for fortran like array
        matlab_array._real = (
            cpx.real.ravel()
        )  # second ravel to correct the strides 18->8
        matlab_array._imag = cpx.imag.ravel()
        matlab_array.reshape(shape)
        # matlab_array._strides=strides

    else:  # real case     # create empty matlab.mlarray
        if numpy_array.dtype == numpy.float64:
            matlab_array = matlab.double(
                initializer=None, size=(1, num_elements), is_complex=False
            )
        elif numpy_array.dtype == numpy.int64:
            matlab_array = matlab.int64(initializer=None, size=(1, num_elements))
        elif numpy_array.dtype == numpy.bool:
            matlab_array = matlab.logical(initializer=None, size=(1, num_elements))
        else:
            raise TypeError(f"Type {numpy_array.dtype} is missing")

        matlab_array._data = numpy_array.ravel(order=order)  # associate the data
        # print(matlab_array._data.flags,matlab_array._data,'\n') # control owner

        matlab_array.reshape(shape)  # back to original shape
        if (
            len(shape) > 1
        ):  # array strides are in number of cell (numpy strides are in bytes)
            # if len(shape)==1 no need to change. Format pb because _stride expect (1,1) and stride = (1,)
            matlab_array._strides = (
                strides  # change stride (matlab use 'F' order ie [nc,1] )
            )

    return matlab_array


def dict_to_sparse(K):
    """Create a scipy sparse CSR matrix from dictionary

    Parameters
    -----------
    K : dictionary K['i'], K['j'] and K['s']
      The sparse matrix in the coo format. K['i'], K['j'] contains the row
      and column index (int64) and the values K['s']  (double or complex)

    Returns
    -------
    Ksp :  sparse.csr_matrix
      The converted sparse matrix. csr is faster for computation."""
    # get shape
    shape = tuple(K["shape"]._data)
    # -1 because matlab index start to 1
    if len(K["i"]._data) > 1:
        ksp = sparse.coo_matrix(
            (
                matlab_to_ndarray(K["s"]).ravel(),
                (
                    matlab_to_ndarray(K["i"]).ravel() - 1,
                    matlab_to_ndarray(K["j"]).ravel() - 1,
                ),
            ),
            shape=shape,
        ).tocsr()
    else:
        raise TypeError(
            "The sparse matrix contains just one element and matlab returns just scalar..."
        )

    return ksp


if __name__ == "__main__":
    a = numpy.random.random((2, 1, 2))
    b = ndarray_to_matlab(a)
    c = matlab_to_ndarray(b)
    print(a.shape, b.size, c.shape)

    print(a)
    print(b)
    print(c)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

from enum import Enum
from typing import Tuple, Union

import numpy

__all__ = [
    "pil_merge_images",
    "pil_img_to_np_array",
    "np_array_to_pil_img",
    "pil_image_to_byte_array",
    "byte_array_to_pil_image",
]

import io

from PIL import Image
from sorcery import assigned_names


class PilModesEnum(Enum):
    """
      PIL pixel formats:

    RGB 24bits per pixel, 8-bit-per-channel RGB), 3 channels
    RGBA (8-bit-per-channel RGBA), 4 channels
    RGBa (8-bit-per-channel RGBA, remultiplied alpha), 4 channels
    1 - 1bpp, often for masks, 1 channel
    L - 8bpp, grayscale, 1 channel
    P - 8bpp, paletted, 1 channel
    I - 32-bit integers, grayscale, 1 channel
    F - 32-bit floats, grayscale, 1 channel
    CMYK - 8 bits per channel, 4 channels
    YCbCr - 8 bits per channel, 3 channels
    """

    OneBpp = "1"
    CMYK, F, HSV, I, L, LAB, P, RGB, RGBA, RGBX, YCbCr = assigned_names()


"""
1 (1-bit pixels, black and white, stored with one pixel per byte)

L (8-bit pixels, black and white)

P (8-bit pixels, mapped to any other mode using a color palette)

RGB (3x8-bit pixels, true color)

RGBA (4x8-bit pixels, true color with transparency mask)

CMYK (4x8-bit pixels, color separation)

YCbCr (3x8-bit pixels, color video format)

Note that this refers to the JPEG, and not the ITU-R BT.2020, standard

LAB (3x8-bit pixels, the L*a*b color space)

HSV (3x8-bit pixels, Hue, Saturation, Value color space)

I (32-bit signed integer pixels)

F (32-bit floating point pixels)

# Pillow also provides limited support for a few additional modes, including:

LA (L with alpha)

PA (P with alpha)

RGBX (true color with padding)

RGBa (true color with premultiplied alpha)

La (L with premultiplied alpha)

I;16 (16-bit unsigned integer pixels)

I;16L (16-bit little endian unsigned integer pixels)

I;16B (16-bit big endian unsigned integer pixels)

I;16N (16-bit native endian unsigned integer pixels)

BGR;15 (15-bit reversed true colour)

BGR;16 (16-bit reversed true colour)

BGR;24 (24-bit reversed true colour)

BGR;32 (32-bit reversed true colour)
"""


def pil_image_to_byte_array(image: Image.Image, *, coding: str = "PNG") -> bytes:
    """
    PNG encoded by default
    :param coding:
    :param image:
    :return:"""
    buffer = io.BytesIO()
    image.save(buffer, coding)
    return buffer.getvalue()


def byte_array_to_pil_image(byte_array: bytes) -> Image.Image:
    """

    :param byte_array:
    :return:"""
    return Image.open(io.BytesIO(byte_array))


def pil_img_to_np_array(
    data_path: Union[str, Image.Image],
    *,
    desired_size: Tuple[int, int] = None,
    expand: int = False
) -> numpy.ndarray:
    """
    Util function for loading RGB image into a numpy array.

    Returns array of shape (1, H, W, C)."""
    img = Image.open(data_path)
    img = img.convert("RGB")
    if desired_size:
        img = img.resize((desired_size[1], desired_size[0]))
    x = numpy.asarray(img, dtype="float32")
    if expand:
        x = numpy.expand_dims(x, axis=0)
    return x / 255.0


def np_array_to_pil_img(x: numpy.ndarray) -> Image.Image:
    """
    Util function for converting a numpy array to a PIL img.

    Returns PIL RGB img."""
    x = numpy.asarray(x)
    x = x + max(-numpy.min(x), 0)
    x_max = numpy.max(x)
    if x_max != 0:
        x /= x_max
    return Image.fromarray((x * 255.0).astype("uint8"), "RGB")


def pil_merge_images(image1: Image.Image, image2: Image.Image) -> Image.Image:
    """Merge two images into one, displayed side by side."""
    (width1, height1) = image1.size
    (width2, height2) = image2.size

    result_width = width1 + width2
    result_height = max(height1, height2)

    result = Image.new("RGB", (result_width, result_height))
    result.paste(im=image1, box=(0, 0))
    result.paste(im=image2, box=(width1, 0))
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 29/07/2020
           """

__all__ = ["plot_grad_flow"]

import numpy
import torch
from matplotlib import pyplot
from matplotlib.lines import Line2D

from draugr.torch_utilities.optimisation.parameters import normal_init_weights


def plot_grad_flow(
    model: torch.nn.Module,
    lines: bool = True,
    alpha: float = 0.5,
    line_width: float = 1.0,
) -> None:
    """

    Plots the gradients flowing through different layers in the net during training.
    Can be used for checking for possible gradient vanishing / exploding problems.

    Usage: After loss.backwards(), use plot_grad_flow(model) to visualize the gradient flow of model

    :param model:
    :type model:
    :param lines:
    :type lines:
    :param alpha:
    :type alpha:
    :param line_width:
    :type line_width:"""
    assert 0.0 < alpha <= 1.0
    ave_grads = []
    max_grads = []
    layers = []

    for n, p in model.named_parameters():
        if p.requires_grad and ("bias" not in n):
            layers.append(n)
            grad_abs = p.grad.abs()
            ave_grads.append(grad_abs.mean())
            max_grads.append(grad_abs.max())

    if lines:
        pyplot.plot(max_grads, alpha=alpha, linewidth=line_width, color="r")
        pyplot.plot(ave_grads, alpha=alpha, linewidth=line_width, color="g")
    else:
        pyplot.bar(
            numpy.arange(len(max_grads)),
            max_grads,
            alpha=alpha,
            linewidth=line_width,
            color="r",
        )
        pyplot.bar(
            numpy.arange(len(max_grads)),
            ave_grads,
            alpha=alpha,
            linewidth=line_width,
            color="g",
        )

    pyplot.hlines(0, 0, len(ave_grads) + 1, linewidth=1, color="k")
    pyplot.xticks(range(0, len(ave_grads), 1), layers, rotation="vertical")
    pyplot.xlim(left=0, right=len(ave_grads))
    max_g = max(max_grads)
    margin = max_g * 1.1
    pyplot.ylim(
        bottom=max_g - margin, top=margin
    )  # zoom in on the lower gradient regions
    pyplot.xlabel("Layers")
    pyplot.ylabel("Gradient Magnitude")
    pyplot.title("Gradient Flow")
    pyplot.grid(True)
    pyplot.legend(
        [
            Line2D([0], [0], color="c", lw=4),
            Line2D([0], [0], color="b", lw=4),
            Line2D([0], [0], color="k", lw=4),
        ],
        ["max-gradient", "mean-gradient", "zero-gradient"],
    )


if __name__ == "__main__":

    def a() -> None:
        """
        :rtype: None
        """
        input = torch.randn(10, 50, requires_grad=True)
        target = torch.empty(10, dtype=torch.long).random_(2)
        model = torch.nn.Sequential(
            torch.nn.Linear(50, 50),
            torch.nn.ReLU(),
            torch.nn.Linear(50, 50),
            torch.nn.ReLU(),
            torch.nn.Linear(50, 50),
            torch.nn.ReLU(),
            torch.nn.Linear(50, 50),
            torch.nn.ReLU(),
            torch.nn.Linear(50, 2),
        )
        normal_init_weights(model, std=1.2)
        criterion = torch.nn.CrossEntropyLoss()
        outputs = model(input)
        loss = criterion(outputs, target)
        loss.backward()
        plot_grad_flow(model)
        pyplot.show()

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import os

__author__ = "Christian Heider Nielsen"

from pathlib import Path
from typing import Dict, Sequence, Iterable, Union

from draugr.numpy_utilities.datasets.splitting import (
    SplitEnum,
    train_valid_test_split,
)
from draugr.numpy_utilities.datasets.defaults import DEFAULT_ACCEPTED_FILE_FORMATS
from warg import drop_unused_kws

__all__ = ["build_shallow_categorical_dataset"]


@drop_unused_kws
def build_shallow_categorical_dataset(
    directory: Union[Path, str],
    *,
    validation_percentage: float = 15,
    testing_percentage: float = 0,
    extensions: Iterable = DEFAULT_ACCEPTED_FILE_FORMATS,
    is_valid_file: callable = None,
    verbose: bool = False,
) -> Dict[str, Dict[SplitEnum, Sequence]]:
    """
    Returns:
    An OrderedDict containing an entry for each label subfolder, with images
    split into training, testing, and validation sets within each label.
    The order of items defines the class indices.
      :param is_valid_file:
      :param directory:
      :param validation_percentage:
      :param testing_percentage:
      :param extensions:
      :param verbose:
      :return:"""

    if not isinstance(directory, Path):
        directory = Path(directory)

    if not extensions:
        extensions = DEFAULT_ACCEPTED_FILE_FORMATS

    if not directory.exists():
        logging.error(f"directory {directory} not found.")
        raise FileNotFoundError(f"directory {directory} not found.")

    categories_dict = {category: [] for category in next(os.walk(str(directory)))[1]}
    logging.info(f"Found categories {categories_dict.keys()}")

    for c in categories_dict.keys():
        for sub_directory in sorted([Path(x[0]) for x in os.walk(str(directory / c))]):
            logging.info(f"Looking for samples in {sub_directory}")
            for extension in sorted(set(os.path.normcase(ext) for ext in extensions)):
                extension = extension.lstrip(".")
                files = list(sub_directory.glob(f"*.{extension}"))
                logging.info(
                    f"Found {len(files)} samples of type {extension} for category {c}"
                )
                categories_dict[c].extend(files)

    if verbose:
        print(categories_dict)

    return train_valid_test_split(
        categories_dict,
        testing_percentage=testing_percentage,
        validation_percentage=validation_percentage,
    )


if __name__ == "__main__":

    def asd() -> None:
        """
        :rtype: None
        """
        from draugr.visualisation import indent_lines
        from draugr.numpy_utilities.datasets.splitting import SplitEnum

        a = build_shallow_categorical_dataset(
            Path.home() / "Data" / "mnist_png" / "training", testing_percentage=0
        )

        for k in a.keys():
            total = (
                len(a[k][SplitEnum.training])
                + len(a[k][SplitEnum.validation])
                + len(a[k][SplitEnum.testing])
            )
            print(f"\n{k}:")
            print(indent_lines(len(a[k][SplitEnum.training]) / total))
            print(indent_lines(len(a[k][SplitEnum.validation]) / total))
            print(indent_lines(len(a[k][SplitEnum.testing]) / total))

    asd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
from typing import Sequence

import numpy
import torch
from draugr.torch_utilities.optimisation.parameters.initialisation import (
    constant_init,
    fan_in_init,
)
from draugr.torch_utilities.tensors.to_tensor import to_tensor
from draugr.torch_utilities.architectures.architecture import Architecture
from draugr.random_utilities import seed_stack

from numpy import prod
from torch import nn
from torch.nn import Module
from warg import NOD

__author__ = "Christian Heider Nielsen"

__doc__ = r"""
Description: Multi Layer Perceptron
Author: Christian Heider Nielsen
"""

__all__ = ["MLP"]


class MLP(Architecture):
    """
    OOOO input_shape
    |XX|                                        fc1
    OOOO hidden_layer_size * (Weights,Biases)
    |XX|                                        fc2
    OOOO hidden_layer_size * (Weights,Biases)
    |XX|                                        fc3
    0000 output_shape * (Weights,Biases)"""

    def __init__(
        self,
        *,
        input_shape: Sequence[int] = None,
        hidden_layers: Sequence[int] = None,
        hidden_layer_activation: Module = torch.nn.ReLU(),
        output_shape: Sequence[int] = None,
        output_activation: Module = torch.nn.Identity(),
        use_bias: bool = True,
        use_dropout: bool = False,
        dropout_prob: float = 0.2,
        auto_build_hidden_layers_if_none=True,
        input_multiplier: int = 32,
        max_layer_width=1000,
        output_multiplier: int = 16,
        default_init: callable = fan_in_init,
        # prefix:str=None, #TODO name sub networks
        **kwargs,
    ):
        assert input_shape is not None
        assert output_shape is not None

        input_shape = self.infer_input_shape(input_shape)
        output_shape = self.infer_output_shape(output_shape)

        super().__init__(input_shape=input_shape, output_shape=output_shape, **kwargs)

        self._use_bias = use_bias

        if not hidden_layers and auto_build_hidden_layers_if_none:
            hidden_layers = self.construct_progressive_hidden_layers(
                self._input_shape,
                self._output_shape,
                input_multiplier,
                output_multiplier,
                max_layer_width=max_layer_width,
            )
        elif isinstance(hidden_layers, int):
            hidden_layers = (hidden_layers,)

        if use_dropout:
            HiddenLinear = lambda *a, **b: nn.Sequential(
                nn.Linear(*a, **b), nn.Dropout(p=dropout_prob), hidden_layer_activation
            )
        else:
            HiddenLinear = lambda *a, **b: nn.Sequential(
                nn.Linear(*a, **b), hidden_layer_activation
            )

        self._hidden_layers = hidden_layers
        self.num_of_layer = len(self._hidden_layers) - 1
        assert self.num_of_layer >= 0

        for i, siz in enumerate(self._input_shape):
            setattr(
                self,
                f"_in{i}",
                HiddenLinear(siz, self._hidden_layers[0], bias=self._use_bias),
            )

        previous_layer_size = self._hidden_layers[0] * len(self._input_shape)

        for i in range(self.num_of_layer):
            setattr(
                self,
                f"_hidden{i}",
                HiddenLinear(
                    previous_layer_size, self._hidden_layers[i], bias=self._use_bias
                ),
            )
            previous_layer_size = self._hidden_layers[i]

        for i, siz in enumerate(self._output_shape):
            setattr(
                self,
                f"_out{i}",
                nn.Sequential(
                    nn.Linear(previous_layer_size, siz, bias=self._use_bias),
                    output_activation,
                ),
            )

        if default_init:
            default_init(self)

    @staticmethod
    def construct_progressive_hidden_layers(
        _input_shape,
        _output_shape,
        input_multiplier: float = 32,
        output_multiplier: float = 16,
        max_layer_width: int = 1000,
    ) -> Sequence[nn.Module]:
        """

        :param _input_shape:
        :type _input_shape:
        :param _output_shape:
        :type _output_shape:
        :param input_multiplier:
        :type input_multiplier:
        :param output_multiplier:
        :type output_multiplier:
        :param max_layer_width:
        :type max_layer_width:
        :return:
        :rtype:
        """
        h_first_size = min(int(sum(_input_shape) * input_multiplier), max_layer_width)
        h_last_size = min(int(sum(_output_shape) * output_multiplier), max_layer_width)

        h_middle_size = int(numpy.sqrt(h_first_size * h_last_size))

        hidden_layers = NOD(h_first_size, h_middle_size, h_last_size).as_list()

        return hidden_layers

    @staticmethod
    def infer_input_shape(input_shape: Sequence[int]) -> Sequence[int]:
        """

            :return:
            :rtype:
        :param input_shape:
        :return:"""
        if isinstance(input_shape, Sequence):
            assert len(input_shape) > 0, f"Got length {len(input_shape)}"
            new_input_shape = input_shape
        elif isinstance(input_shape, int):
            new_input_shape = (input_shape,)
            logging.info(
                f"Inflating input shape {input_shape} to vectorised input shape {new_input_shape}"
            )
        else:
            raise ValueError(f"Can not use {input_shape} as input shape")
        return new_input_shape

    @staticmethod
    def infer_output_shape(output_shape: Sequence[int]) -> Sequence[int]:
        """

            :return:
            :rtype:
        :param output_shape:
        :return:"""
        if isinstance(output_shape, Sequence):
            assert len(output_shape) > 0, f"Got length {len(output_shape)}"
            new_output_shape = output_shape
        elif isinstance(output_shape, int):
            new_output_shape = (output_shape,)
            logging.info(
                f"Inflating output shape {output_shape} to vectorised output shape {new_output_shape}"
            )
        else:
            raise ValueError(f"Can not use {output_shape} as output shape")

        return new_output_shape

    def forward(self, *x, **kwargs):
        """

        :param x:
        :return output:"""
        if len(x) != len(self.input_shape):
            raise ValueError(
                f"{len(self.input_shape)} input arguments expected, {len(x)} was supplied"
            )

        ins = []
        for i in range(len(self._input_shape)):
            ins.append(getattr(self, f"_in{i}")(x[i]))

        val = torch.cat(ins, dim=-1)
        for i in range(self.num_of_layer):
            val = getattr(self, f"_hidden{i}")(val)

        outs = []
        for i in range(len(self._output_shape)):
            outs.append(getattr(self, f"_out{i}")(val))

        if len(outs) == 1:
            return outs[0]

        return outs


if __name__ == "__main__":
    seed_stack(4)

    def stest_single_dim():
        """ """
        pos_size = (4,)
        a_size = (1,)
        model = MLP(input_shape=pos_size, output_shape=a_size)

        pos_1 = to_tensor(
            numpy.random.rand(64, pos_size[0]), device="cpu", dtype=torch.float
        )
        print(model(pos_1)[0].shape)

    def stest_hidden_dim():
        """ """
        pos_size = (3,)
        hidden_size = list(range(6, 10))
        a_size = (4,)
        model = MLP(
            input_shape=pos_size,
            hidden_layers=hidden_size,
            output_shape=a_size,
            hidden_layer_activation=torch.nn.Tanh(),
            default_init=None,
        )

        model2 = nn.Sequential(
            *[
                nn.Linear(3, 6),
                nn.Tanh(),
                nn.Linear(6, 7),
                nn.Tanh(),
                nn.Linear(7, 8),
                nn.Tanh(),
                nn.Linear(8, 9),
                nn.Tanh(),
                nn.Linear(9, 4),
            ]
        )
        model3 = nn.Sequential(
            *[
                nn.Linear(3, 6),
                nn.Tanh(),
                nn.Linear(6, 7),
                nn.Tanh(),
                nn.Linear(7, 8),
                nn.Tanh(),
                nn.Linear(8, 9),
                nn.Tanh(),
                nn.Linear(9, 4),
            ]
        )
        constant_init(model, 0.142)
        constant_init(model2, 0.142)
        constant_init(model3, 0.142)
        print(model, model2, model3)

        pos_1 = to_tensor(
            numpy.random.rand(64, pos_size[0]), device="cpu", dtype=torch.float
        )
        print(model(pos_1)[0].shape)
        print(model2(pos_1).shape)
        print(model3(pos_1).shape)

    def stest_multi_dim_in():
        """ """
        pos_size = (2, 3, 2)
        a_size = (2, 4, 5)
        model = MLP(input_shape=pos_size, output_shape=a_size)

        pos_1 = to_tensor(
            numpy.random.rand(64, prod(pos_size[1:])), device="cpu", dtype=torch.float
        )
        pos_2 = to_tensor(
            numpy.random.rand(64, prod(pos_size[1:])), device="cpu", dtype=torch.float
        )
        print(model(pos_1, pos_2)[0].shape)

    def stest_multi_dim_out():
        """ """
        pos_size = (10,)
        a_size = (2, 1)
        model = MLP(input_shape=pos_size, hidden_layers=(100,), output_shape=a_size)

        pos_1 = to_tensor(
            numpy.random.rand(64, *pos_size), device="cpu", dtype=torch.float
        )
        res = model(pos_1)
        print(model)
        print(len(res), res[0].shape, res[1].shape)

    def stest_multi_dim_both():
        """ """
        pos_size = (2, 3)
        a_size = (2, 4, 5)
        model = MLP(input_shape=pos_size, output_shape=a_size)

        pos_1 = to_tensor(
            numpy.random.rand(64, pos_size[0]), device="cpu", dtype=torch.float
        )
        pos_2 = to_tensor(
            numpy.random.rand(64, pos_size[1]), device="cpu", dtype=torch.float
        )
        res = model(pos_1, pos_2)
        print(model)
        print(len(res), res[0].shape, res[1].shape, res[2].shape)

    def stest_auto():
        """ """
        pos_size = (4,)
        a_size = (2,)
        model = MLP(input_shape=pos_size, output_shape=a_size)

        pos_1 = to_tensor(
            numpy.random.rand(64, pos_size[0]), device="cpu", dtype=torch.float
        )
        res = model(pos_1)
        print(model)
        print(len(res), res[0].shape)

    # stest_single_dim()
    # stest_hidden_dim()
    # stest_multi_dim_both()
    # stest_multi_dim_out()
    stest_auto()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 29/07/2020
           """

__all__ = ["register_bad_grad_hooks", "print_grad_trace"]

from typing import Any

import torch


def register_bad_grad_hooks(var: Any) -> callable:
    """

    :param var:
    :type var:
    :return:
    :rtype:"""

    def iter_graph(root, callback):
        """

        :param root:
        :type root:
        :param callback:
        :type callback:"""
        queue = [root]
        seen = set()
        while queue:
            fn = queue.pop()
            if fn in seen:
                continue
            seen.add(fn)
            for next_fn, _ in fn.next_functions:
                if next_fn is not None:
                    queue.append(next_fn)
            callback(fn)

    fn_dict = {}

    def hook_callback(fn):
        """

        :param fn:
        :type fn:"""

        def register_grad(grad_input, grad_output):
            """

            :param grad_input:
            :type grad_input:
            :param grad_output:
            :type grad_output:"""
            fn_dict[fn] = grad_input

        fn.register_hook(register_grad)

    iter_graph(var.grad_fn, hook_callback)

    def is_bad_grad(grad_output):
        """

        :param grad_output:
        :type grad_output:
        :return:
        :rtype:"""
        if grad_output is None:
            return False
        return torch.isnan(grad_output).any() or (grad_output.abs() >= 1e6).any()

    def make_dot():
        """

        :return:
        :rtype:"""
        from graphviz import Digraph

        node_attr = dict(
            style="filled",
            shape="box",
            align="left",
            fontsize="12",
            ranksep="0.1",
            height="0.2",
        )
        dot_ = Digraph(node_attr=node_attr, graph_attr=dict(size="12,12"))

        def size_to_str(size):
            """

            :param size:
            :type size:
            :return:
            :rtype:"""
            return f'({", ".join(map(str, size))})'

        def build_graph(fn):
            """

            :param fn:
            :type fn:"""
            if hasattr(fn, "variable"):  # if GradAccumulator
                u = fn.variable
                node_name = "Variable\n " + size_to_str(u.size())
                dot_.node(str(id(u)), node_name, fillcolor="lightblue")
            else:
                assert fn in fn_dict, fn
                fillcolor = "white"
                if any(is_bad_grad(gi) for gi in fn_dict[fn]):
                    fillcolor = "red"
                dot_.node(str(id(fn)), str(type(fn).__name__), fillcolor=fillcolor)
            for next_fn, _ in fn.next_functions:
                if next_fn is not None:
                    next_id = id(getattr(next_fn, "variable", next_fn))
                    dot_.edge(str(next_id), str(id(fn)))

        iter_graph(var.grad_fn, build_graph)

        return dot_

    return make_dot


def print_grad_trace(var_grad_fn):
    """ """
    print(var_grad_fn)
    for n in var_grad_fn.next_functions:
        if n[0]:
            try:
                tensor = getattr(n[0], "variable")
                print(n[0])
                print("Tensor with grad found:", tensor)
                print(" - gradient:", tensor.grad)
                print()
            except AttributeError as e:
                print_grad_trace(n[0])


if __name__ == "__main__":

    def asdifiejsf() -> None:
        """
        :rtype: None
        """
        x = torch.randn(10, 10, requires_grad=True)
        y = torch.randn(10, 10, requires_grad=True)

        z = x / (y * 0)
        z = z.sum() * 2
        get_dot = register_bad_grad_hooks(z)
        z.backward()
        dot = get_dot()
        # dot.save('tmp.dot') # to get .dot
        # dot.render('tmp') # to get SVG

    asdifiejsf()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 22/06/2020
           """

__all__ = ["MATLAB_ENGINE", "start_engine"]

from typing import Any

MATLAB_ENGINE = None


def start_engine() -> Any:
    """ """
    return
    global MATLAB_ENGINE
    if not MATLAB_ENGINE:
        from matlab import engine

        print("Starting Matlab")
        if True:
            MATLAB_ENGINE = engine.start_matlab()
        else:
            future = engine.start_matlab(background=True)
            MATLAB_ENGINE = future.result()
        print("Matlab started")
    else:
        print("Matlab already running")

    return MATLAB_ENGINE


if __name__ == "__main__":
    engine = start_engine()
    print(engine)
    print(MATLAB_ENGINE)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-06-2021
           """

__all__ = ["to_latex_table"]

from torch import nn


def to_latex_table(m: nn.Module) -> str:
    """

    :param m:
    :return:
    """
    return m.__repr__()


if __name__ == "__main__":
    from torch.nn import functional

    def aiasujd():
        """

        :return:
        """

        class Model(nn.Module):
            """ """

            def __init__(self):
                super().__init__()

                self.conv0 = nn.Conv2d(1, 16, kernel_size=3, padding=5)
                self.conv1 = nn.Conv2d(16, 32, kernel_size=3)

            def forward(self, x):
                """

                :param x:
                :return:
                """
                h = self.conv0(x)
                h = self.conv1(h)
                return h

        model = Model()

        print(to_latex_table(model))

    def uiahsduhaw():
        """

        :return:
        """

        class Model(nn.Module):
            """ """

            def __init__(self):
                super().__init__()
                self.conv1 = nn.Conv2d(1, 10, kernel_size=5)
                self.conv2 = nn.Conv2d(10, 20, kernel_size=5)
                self.conv2_drop = nn.Dropout2d()
                self.fc1 = nn.Linear(320, 50)
                self.fc2 = nn.Linear(50, 10)

            def forward(self, x):
                """

                :param x:
                :return:
                """
                x = functional.relu(functional.max_pool2d(self.conv1(x), 2))
                x = functional.relu(
                    functional.max_pool2d(self.conv2_drop(self.conv2(x)), 2)
                )
                x = x.view(-1, 320)
                x = functional.relu(self.fc1(x))
                x = functional.dropout(x, training=self.training)
                x = self.fc2(x)
                return functional.log_softmax(x, dim=1)

        model = Model()

        print(to_latex_table(model))

    uiahsduhaw()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

import statistics

__all__ = ["MetricSummary"]


class MetricSummary:
    """ """

    def __init__(self):
        self._values = []
        self.length = 0
        self.running_mean = 0
        self.running_variance = 0

    def append(self, val):
        """

        :param val:
        :type val:"""
        self._values.append(val)
        self.length += 1

    @property
    def values(self):
        """

        :return:
        :rtype:"""
        return self._values

    def moving_average(self, window_size=100):
        """

        :param window_size:
        :type window_size:
        :return:
        :rtype:"""
        if self.length >= window_size:
            return statistics.mean(self._values[-window_size:])
        elif self.length > 0:
            return statistics.mean(self._values)
        else:
            return 0

    def running_average(self, data):
        """
        Computes new running mean and variances.
        :param data: New piece of data.
        :return: New mean and variance values."""
        mean, var, steps = self.running_mean, self.running_variance, self.length
        current_x = statistics.mean(data)

        new_mean = mean + (current_x - mean) / (steps + 1)
        new_variance = var + (current_x - new_mean) * (current_x - mean)

        self.running_mean, self.running_variance = new_mean, new_variance
        return new_mean, new_variance

    def __len__(self):
        return len(self._values)


if __name__ == "__main__":
    MetricSummary()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from pathlib import Path
from typing import List
from warnings import warn

__author__ = "Christian Heider Nielsen"

import statistics

__all__ = ["MetricAggregator", "save_metric"]


class MetricAggregator(object):
    """ """

    def __init__(
        self,
        measures=statistics.__all__[1:],
        keep_measure_history=False,
        use_disk_cache=True,
    ):
        self._values = []
        self._length = 0

        self._running_value = None
        self._running_value_key = "running_value"

        # for key in self._measure_keys:
        #  setattr(self,key,None)

        self._stat_measure_keys = measures
        self._keep_measure_history = keep_measure_history
        if self._keep_measure_history:
            self._measures = {}
            for key in self._stat_measure_keys:
                self._measures[key] = []
            self._measures[self._running_value_key] = []

    @property
    def values(self):
        """

        :return:
        :rtype:"""
        return self._values

    @property
    def max(self):
        """

        :return:
        :rtype:"""
        return max(self._values)

    @property
    def min(self):
        """

        :return:
        :rtype:"""
        return min(self._values)

    @property
    def measures(self):
        """

        :return:
        :rtype:"""
        if self._keep_measure_history:
            return self._measures
        else:
            out = {}
            for key in self._stat_measure_keys:
                try:
                    val = getattr(statistics, key)(self._values)
                except statistics.StatisticsError as e:
                    # TODO: warn(f'{e}')
                    val = None
                out[key] = val
            return out

    def add(self, values):
        """

        :param values:
        :type values:"""
        self.append(values)

    def append(self, values):
        """

        :param values:
        :type values:"""
        self._values.append(values)
        if type is list:
            self._length += len(values)
        else:
            self._length += 1

        self.calc_running_value(values)

        if self._keep_measure_history:
            for key in self._stat_measure_keys:
                if self._length > 1:
                    try:
                        val = getattr(statistics, key)(self._values)
                    except:
                        val = None
                    self._measures[key].append(val)
                else:
                    # warn(f'Length of statistical values are <=1, measure "{key}" maybe ill-defined')
                    try:
                        val = getattr(statistics, key)(self._values)
                    except statistics.StatisticsError as e:
                        # TODO: warn(f'{e}')
                        val = None
                    self._measures[key].append(val)

    # def __setitem__(self, key, value):
    #    if self._keep_measure_history:
    #        self.

    def __getitem__(self, item):
        return self._values[item]

    def __contains__(self, item):
        return self._values[item]

    def __iter__(self):
        return self._values

    def __getattr__(self, item):
        if item in self._stat_measure_keys:
            if self._length > 1:
                if self._keep_measure_history:
                    return self._measures[item]
                else:
                    try:
                        return getattr(statistics, item)(self._values)
                    except statistics.StatisticsError as e:
                        warn(f"{e}")
                        return None
            else:
                warn(
                    f'Length of statistical values are <=1, measure "{item}" maybe ill-defined'
                )
                try:
                    return getattr(statistics, item)(self._values)
                except statistics.StatisticsError as e:
                    warn(f"{e}")
                    return None
        elif item == self._running_value_key:
            return self._measures[item]
        else:
            raise AttributeError

    # def __call__(self, *args, **kwargs):
    #  return self._values

    def __repr__(self):
        return f"<StatisticAggregator> values: {self._values}, measures: {self.measures} </StatisticAggregator>"

    def __str__(self):
        return str(self._values)

    def __len__(self):
        return len(self._values)

    def calc_moving_average(self, window_size=100):
        """

        :param window_size:
        :type window_size:
        :return:
        :rtype:"""
        if self._length >= window_size:
            return statistics.mean(self._values[-window_size:])
        elif self._length > 0:
            return statistics.mean(self._values)
        else:
            return 0

    def calc_running_value(self, new_val=None, *, lambd=0.99):
        """

        :param new_val:
        :type new_val:
        :param lambd:
        :type lambd:
        :return:
        :rtype:"""
        if new_val is None:
            return self._running_value

        if new_val is list:
            for nw in new_val:
                self.calc_running_value(nw, lambd=lambd)

        if self._running_value:
            self._running_value = self._running_value * lambd + new_val * (1 - lambd)
        else:
            self._running_value = new_val

        if self._keep_measure_history:
            self._measures[self._running_value_key].append(self._running_value)

        return self._running_value

    def save(
        self, *, stat_name, project_name="non", config_name="non", directory="logs"
    ):
        """

        :param stat_name:
        :type stat_name:
        :param project_name:
        :type project_name:
        :param config_name:
        :type config_name:
        :param directory:
        :type directory:"""
        save_metric(
            self._values,
            metric_name=stat_name,
            project_name=project_name,
            config_name=config_name,
            directory=directory,
        )


def save_metric(
    metric: List[MetricAggregator],
    *,
    metric_name,
    project_name,
    config_name,
    directory=Path("logs"),
) -> bool:
    """

    :param metric:
    :type metric:
    :param metric_name:
    :type metric_name:
    :param project_name:
    :type project_name:
    :param config_name:
    :type config_name:
    :param directory:
    :type directory:
    :return:
    :rtype:"""
    import csv
    import datetime

    if metric:
        _file_date = datetime.datetime.now()
        _file_name = (
            f'{project_name}-{config_name.replace(".", "_")}-'
            f'{_file_date.strftime("%y%m%d%H%M")}.{metric_name}.csv'
        )
        _file_path = directory / _file_name

        stat = [[s] for s in metric]
        with open(_file_path, "w") as f:
            w = csv.writer(f)
            w.writerows(stat)
        print(f"Saved metric at {_file_path}")

        return True
    return False


if __name__ == "__main__":
    signals = MetricAggregator(keep_measure_history=False)

    for i in range(10):
        signals.append(i)

    print(signals)
    print(signals.measures)
    print(signals.variance)
    print(signals.calc_moving_average())
    print(signals.max)
    print(signals.min)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04-03-2021
           """

from typing import Tuple

from draugr.scipy_utilities import mag_decimation_subsample

__all__ = [
    "annotate_lines",
    "default_index_decimator",
]

from warg import GDKC, passes_kws_to

from matplotlib.pyplot import Axes

default_index_decimator = GDKC(
    mag_decimation_subsample, decimation_factor=5, return_indices=True
)  # finds interesting features?


@passes_kws_to(Axes.annotate)
def annotate_lines(
    ax_: Axes,
    num_lines: int = 1,  # None for all
    index_decimator: callable = default_index_decimator,
    color: str = "k",  # None for auto color
    xycoords: Tuple[str, str] = (
        "data",
        # 'axes fraction',
        "data",
    ),  # TODO: NOT DONE! Where to place annotation, use 'axes fraction' for along axes'
    ha: str = "left",
    va: str = "center",
    **kwargs,
) -> None:
    """

    :param ax_:
    :param num_lines:
    :param index_decimator:
    :param color:
    :param xycoords:
    :param ha:
    :param va:
    :param kwargs:
    """
    lines = ax_.lines
    if not num_lines:
        num_lines = len(lines)
    for l, _ in zip(lines, range(num_lines)):
        y = l.get_ydata()
        x = l.get_xdata()

        if not color:
            color = l.get_color()

        if index_decimator:
            mag_y = index_decimator(y)
        else:
            mag_y = list(range(len(y)))

        for x_, y_ in zip(x[mag_y], y[mag_y]):
            ax_.annotate(
                f"{y_:.2f}",
                xy=(x_, y_),  # ( 1, y_) axes fraction'
                xycoords=xycoords,
                ha=ha,
                va=va,
                color=color,
                **kwargs,
            )


if __name__ == "__main__":

    def hsdh() -> None:
        """
        :rtype: None
        """
        from matplotlib import pyplot

        a = [*range(0, 10), *range(10, -10, -1), *range(-10, 0)]
        ax_ = pyplot.plot(a)
        annotate_lines(pyplot.gca(), index_decimator=default_index_decimator)
        pyplot.show()

    hsdh()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-05-2021
           """

__all__ = ["pandas_mean_std_bold_formatter"]


def pandas_mean_std_bold_formatter(entry, value, *, precision: int = 3):
    """Format a number in bold when (almost) identical to a given value.

    Args:
        entry: Input number.

        value: Value to compare x with.

        num_decimals: Number of decimals to use for output format.

    Returns:
        String converted output.
        :param entry:
        :param value:
        :param precision:

    """
    # Consider values equal, when rounded results are equal
    # otherwise, it may look surprising in the table where they seem identical
    mean_entry = float(entry.split("\pm")[0])

    if value is not None and round(mean_entry, precision) == round(value, precision):
        return f"$\\mathbf{{{entry}}}$"
    else:
        return f"${entry}$"


def pandas_mean_std_color_formatter(
    entry, value, color: str = "red", *, precision: int = 3
):  # TODO: finish
    """Format a number in bold when (almost) identical to a given value.

    Args:
        entry: Input number.

        value: Value to compare x with.

        num_decimals: Number of decimals to use for output format.

    Returns:
        String converted output.
        :param entry:
        :param value:
        :param color:
        :param precision:

    """
    # Consider values equal, when rounded results are equal
    # otherwise, it may look surprising in the table where they seem identical
    mean_entry = float(entry.split("\pm")[0])

    if value is not None and round(mean_entry, precision) == round(value, precision):
        return f"$\\textcolor{{{color}}}{{{entry}}}$"
    else:
        return f"${entry}$"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-01-2021
           """

from typing import Sequence

import cv2

from draugr.drawers.drawer import Drawer
from draugr.opencv_utilities import WindowFlagEnum
from draugr.opencv_utilities.windows.default import match_return_code
from warg import drop_unused_kws, passes_kws_to

__all__ = ["OpencvImageStream"]


class OpencvImageStream(Drawer):
    """ """

    @drop_unused_kws
    @passes_kws_to(Drawer.__init__)
    def __init__(self, title: str = "", render: bool = True, **kwargs):

        super().__init__(**kwargs)
        if not render:
            return

        self.window_id = title
        cv2.namedWindow(self.window_id, WindowFlagEnum.normal.value)

    def draw(self, data: Sequence):
        """

        :param data:
        :return:"""
        if match_return_code(cv2.waitKey(1)):  # esc to quit
            cv2.destroyWindow(self.window_id)
            raise StopIteration
        else:
            cv2.imshow(self.window_id, data)


if __name__ == "__main__":

    def asdasf() -> None:
        """
        :rtype: None
        """
        from draugr.opencv_utilities import frame_generator, AsyncVideoStream
        from draugr.tqdm_utilities import progress_bar

        with AsyncVideoStream() as vc:
            with OpencvImageStream() as s:
                for i in progress_bar(frame_generator(vc), auto_total_generator=False):
                    s.draw(i)

    asdasf()



import os
import subprocess
from pathlib import Path
from typing import Optional

from apppath import ensure_existence
from warg import Number

FORMAT_LIST = [".mp4", ".avi", ".mkv", ".flv", ".mov"]
AUDIO_FORMAT = ".aac"


__all__ = ["extract_frames"]


def extract_frames(
    file_path: Path,
    frame_out_dir: Optional[Path] = None,
    audio_out_dir: Optional[Path] = None,
    rate: Number = 25,
    frame_format: str = "jpg",
    extract_sound: bool = True,
    ffmpeg_path: Path = "ffmpeg",
):

    root_dir = file_path.parent
    if frame_out_dir is None:
        frame_out_dir = ensure_existence(root_dir / "frames")
    if audio_out_dir is None:
        audio_out_dir = ensure_existence(frame_out_dir.parent / "audio")

    if file_path.is_file() and file_path.suffix in FORMAT_LIST:
        print(f"start extracting {file_path} frames")
        subprocess.call(
            [
                str(ffmpeg_path),
                "-i",
                file_path,
                "-r",
                str(rate),
                "-f",
                "image2",
                "-y",
                "-qscale:v",
                "2",
                str(frame_out_dir / f"{file_path.name}-%05d.{frame_format}"),
            ]
        )
        print(f"end extracting {file_path} frames")

        if extract_sound:
            print(f"start extracting {file_path} audio")
            subprocess.call(
                [
                    str(ffmpeg_path),
                    "-i",
                    file_path,
                    "-vn",
                    "-acodec",
                    "copy",
                    "-y",
                    str(audio_out_dir / f"{file_path.name}{AUDIO_FORMAT}"),
                ]
            )
            print(f"end extracting {file_path} audio")
    else:
        print(f"{file_path} is not a video file")


if __name__ == "__main__":
    extract_frames(
        Path.home() / "Downloads" / "brandt.mp4",
        ffmpeg_path=Path.home()
        / "Downloads"
        / "ffmpeg-5.0-essentials_build"
        / "ffmpeg-5.0-essentials_build"
        / "bin"
        / "ffmpeg.exe",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Union

import numpy
import torch

from draugr.torch_utilities.tensors import to_tensor

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["torch_pi"]


def torch_pi(device: Union[str, torch.device] = "cpu") -> torch.tensor:
    """
    Returns numpy.pi as a tensor

    :param device:
    :type device:
    :return:
    :rtype:"""
    return to_tensor([numpy.math.pi], device=device)



import enum
from enum import Enum
from itertools import zip_longest
from pathlib import Path
from pickle import dump
from typing import Iterable, Mapping, Tuple, TypeVar, Union

import numpy
import pandas
import tensorflow
from PIL.Image import Image
from matplotlib import pyplot
from tensorboard.backend.event_processing import event_accumulator

from apppath import AppPath, ensure_existence

__all__ = ["TensorboardEventExporter"]

from warg import passes_kws_to

# TODO: implement export options using ExportMethodEnum
# TODO: MAJOR REFACTOR INCOMING

TagTypeEnum = TypeVar("TagTypeEnum")


class TensorboardEventExporter:
    """
    Reads event files and exports the requested tags."""

    # class TagTypeEnum(Enum): #Static version, does not adapt to plugins!
    #    images = 'images'
    #    scalars = 'scalars'
    #    tensors = 'tensors'
    #    audio = 'audio'
    #    distributions= 'distributions'
    #    graph = 'graph'
    #    meta_graph = 'meta_graph'
    #    histograms = 'histograms'
    #    run_metadata = 'run_metadata'

    def __init__(
        self,
        path_to_events_file_s: Path,
        size_guidance: Mapping = None,
        *,
        save_to_disk: bool = False,
    ):
        """

        :param path_to_events_file_s:
        :param size_guidance:
        :param save_to_disk:"""
        if size_guidance is None:
            size_guidance = 0

        if isinstance(size_guidance, Mapping):
            pass
        elif isinstance(
            size_guidance, int
        ):  # if only an integer was provided override all size_guidance entries to that integer
            size_guidance_map = (
                event_accumulator.STORE_EVERYTHING_SIZE_GUIDANCE
            )  # Get entries with store everything
            if (
                size_guidance > 0
            ):  # the provided integer was above 0 (store_everything) then limit store to that integer (limited)
                size_guidance_map = {k: size_guidance for k in size_guidance_map.keys()}
            size_guidance = size_guidance_map
        else:
            raise TypeError(f"Invalid type of size guidance {type(size_guidance)}")

        self.path_to_events_file = str(path_to_events_file_s)

        self.event_acc = event_accumulator.EventAccumulator(
            self.path_to_events_file, size_guidance=size_guidance
        )

        self.event_acc.Reload()
        self.tags_available = self.event_acc.Tags()
        self.save_to_disk = save_to_disk

        tags_dict = {}
        for (
            t
        ) in self.tags_available:  # TODO: Automatic but not nice for code completion
            setattr(self, f"available_{t}", self.tags_available[t])
            tags_dict[str(t)] = str(t)

        TensorboardEventExporter.TagTypeEnum = enum.Enum(
            "TagTypeEnum", tags_dict
        )  # dynamic version

    def tag_test(self, *tags, type_str: Union[str, TagTypeEnum]) -> bool:
        """

        :param tags:
        :param type_str:
        :return:"""
        if not len(tags):
            print("No tags requested")
            # raise Exception #TODO: maybe

        if isinstance(type_str, Enum):
            type_str = type_str.value

        if (
            len(tags) == 1
            and isinstance(tags[0], Iterable)
            and not isinstance(tags[0], str)
        ):
            tags = tags[0]
        tags_available = self.tags_available[type_str]
        assert all(
            [tags_available.__contains__(t) for t in tags]
        ), f"{type_str} tags available: {tags_available}, tags requested {tags}"
        return True

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def export_line_plot(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Tuple[pyplot.Figure]:
        """

        :param tags:
        :param out_dir:
        :return:"""
        self.tag_test(*tags, type_str="scalars")
        out = []
        for t in tags:
            w_times, step_nums, vals = zip(*self.event_acc.Scalars(t))
            fig, ax = pyplot.subplots(nrows=1, ncols=1)
            ax.plot(step_nums, vals)
            if self.save_to_disk:
                fig.savefig(str(out_dir / f"{t}_line_plot.png"))
            out.append(ax)
        return (*out,)

    def export_image(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Tuple[Image]:
        """

        :param tags:
        :param out_dir:
        :return:"""
        self.tag_test(*tags, type_str="images")
        out = []
        for t in tags:
            img = self.event_acc.Images(t)  # TODO: CHECK VIDEO compatibility(gifs)
            if self.save_to_disk:
                with open(str(out_dir / f"{t}_img_{img.step}.png"), "wb") as f:
                    f.write(img.encoded_image_string)
            out.append(Image.fromstring(img.encoded_image_string))
        return (*out,)

    def export_scalar(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Iterable:
        """
        if save to files it pickles tags values with file ending .pkl

        :param tags:
        :param out_dir:
        :return:"""
        self.tag_test(*tags, type_str="scalars")
        out = []
        for t in tags:
            w_times, step_nums, vals = zip(*self.event_acc.Scalars(t))
            if self.save_to_disk:
                with open(str(out_dir / f"{t}.pkl"), "wb") as f:
                    dump(vals, f)
            out.append(vals)
        return (*out,)

    def export_distribution(self, *tags: Iterable[str], out_dir: Path = Path.cwd()):
        """

        :param tags:
        :param out_dir:"""
        self.tag_test(*tags, type_str="distributions")
        raise NotImplemented("not implemented yet!")
        out = []
        for t in tags:
            vals = None  # t
            out.append(vals)
        return (*out,)

    def export_tensor(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Iterable:
        """

        :param tags:
        :param out_dir:
        :return:"""
        self.tag_test(*tags, type_str="tensors")
        out = []
        for t in tags:
            w_times, step_nums, vals = zip(*self.event_acc.Tensors(t))
            if self.save_to_disk:
                with open(str(out_dir / f"{t}.pkl"), "wb") as f:
                    dump(vals, f)
            out.append(vals)
        return (*out,)

    def export_graph(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Iterable:
        """

        :param tags:
        :param out_dir:
        :return:"""
        out = []
        w_times, step_nums, vals = zip(*self.event_acc.Graph())
        if self.save_to_disk:
            with open(str(out_dir / f"graph.pkl"), "wb") as f:
                dump(vals, f)
        out.append(vals)
        return (*out,)

    def export_audio(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Iterable:
        """

        :param tags:
        :param out_dir:
        :return:"""
        self.tag_test(*tags, type_str="audio")
        out = []
        for t in tags:
            w_times, step_nums, vals = zip(*self.event_acc.Audio(t))
            if self.save_to_disk:
                with open(str(out_dir / f"{t}.pkl"), "wb") as f:
                    dump(vals, f)
            out.append(vals)
        return (*out,)

    def export_histogram(
        self, *tags: Iterable[str], out_dir: Path = Path.cwd()
    ) -> Iterable:
        """

        https://www.tensorflow.org/api_docs/python/tf/summary/histogram

        :param tags:
        :param out_dir:
        :return:"""
        self.tag_test(*tags, type_str="histograms")

        out = []
        for t in tags:
            w_times, step_nums, vals = zip(*self.event_acc.Histograms(t))
            if self.save_to_disk:
                with open(str(out_dir / f"{t}.pkl"), "wb") as f:
                    dump(vals, f)
            out.append(vals)
        return (*out,)

    @passes_kws_to(pandas.DataFrame.to_csv)
    def scalar_export_csv(
        self,
        *tags: Iterable[str],
        out_dir: Path = Path.cwd(),
        index_label: str = "epoch",
        **kwargs,
    ) -> Tuple[pandas.DataFrame]:
        """
        size_guidance = 0 means all events, no aggregation or dropping

            :param index_label:
        :return:
        :param tags:
        :param out_dir:"""
        if not len(tags):
            print("No tags requested")  # TODO: maybe just return
            # return tuple()

        self.tag_test(*tags, type_str="scalars")

        out = []

        df = pandas.DataFrame(
            list(
                zip_longest(
                    *[
                        list(zip_longest(*self.event_acc.Scalars(t), fillvalue=None))[
                            -1
                        ]
                        for t in tags
                    ],
                    fillvalue=None,
                )
            ),
            columns=tags,
        )
        if self.save_to_disk:
            df.to_csv(
                str(out_dir / f'scalars_{"_".join(tags) if len(tags) else "none"}.csv'),
                columns=tags,
                index_label=index_label,
                **kwargs,
            )
        out.append(df)
        return (*out,)

    @passes_kws_to(pandas.DataFrame.to_csv)
    def pr_curve_export_csv(
        self,
        *tags: Iterable[str],
        out_dir: Path = Path.cwd(),
        index_label: str = "epoch",
        **kwargs,
    ) -> Tuple[pandas.DataFrame]:
        """
        #TODO only supports a single step and tag for now

        size_guidance = 0 means all events, no aggregation or dropping

            :param index_label:
        :return:
        :param tags:
        :param out_dir:"""
        if not len(tags):
            print("No tags requested")  # TODO: maybe just return
            # return tuple()

        self.tag_test(*tags, type_str="tensors")

        out = []

        numpy_rep = numpy.array(
            [
                tensorflow.make_ndarray(b)
                for a in list(
                    zip_longest(
                        *[
                            list(
                                zip_longest(*self.event_acc.Tensors(t), fillvalue=None)
                            )[-1]
                            for t in tags
                        ],
                        fillvalue=None,
                    )
                )
                for b in a
            ]
        )
        labels = (
            "true_positive_counts",
            "false_positive_counts",
            "true_negative_counts",
            "false_negative_counts",
            "precision",
            "recall",
        )
        df = pandas.DataFrame(
            numpy_rep.tolist(), columns=[f"{t}_{l}" for t in tags for l in labels]
        )
        if self.save_to_disk:
            df.to_csv(
                str(out_dir / f'tensors_{"_".join(tags) if len(tags) else "none"}.csv'),
                # columns=tags,
                index_label=index_label,
                **kwargs,
            )
        out.append(df)
        return (*out,)

    @passes_kws_to(pandas.DataFrame.to_csv)
    def tensor_export_csv(
        self,
        *tags: Iterable[str],
        out_dir: Path = Path.cwd(),
        index_label: str = "epoch",
        **kwargs,
    ) -> Tuple[pandas.DataFrame]:
        """

        size_guidance = 0 means all events, no aggregation or dropping

            :param index_label:
        :return:
        :param tags:
        :param out_dir:"""
        if not len(tags):
            print("No tags requested")  # TODO: maybe just return
            # return tuple()

        self.tag_test(*tags, type_str="tensors")

        out = []

        numpy_rep = numpy.array(
            [
                tensorflow.make_ndarray(b)
                for a in list(
                    zip_longest(
                        *[
                            list(
                                zip_longest(*self.event_acc.Tensors(t), fillvalue=None)
                            )[-1]
                            for t in tags
                        ],
                        fillvalue=None,
                    )
                )
                for b in a
            ]
        )

        df = pandas.DataFrame([[numpy_rep.tolist()]], columns=tags)
        if self.save_to_disk:
            df.to_csv(
                str(out_dir / f'tensors_{"_".join(tags) if len(tags) else "none"}.csv'),
                # columns=tags,
                index_label=index_label,
                **kwargs,
            )
        out.append(df)
        return (*out,)


if __name__ == "__main__":

    def a() -> None:
        """
        :rtype: None
        """
        _path_to_events_file = next(
            AppPath("Draugr", "Christian Heider Nielsen").user_log.rglob(
                "events.out.tfevents.*"
            )
        )
        print(_path_to_events_file)
        tee = TensorboardEventExporter(_path_to_events_file.parent, save_to_disk=True)
        print(tee.tags_available)
        # tee.export_csv('train_loss')
        # tee.export_line_plot('train_loss')
        # pyplot.show()
        print(tee.export_histogram())
        print(tee.available_scalars)
        print(
            tee.pr_curve_export_csv(
                *tee.tags_available["tensors"],
                out_dir=ensure_existence(Path.cwd() / "exclude"),
            )
        )
        print(list(iter(tee.TagTypeEnum)))

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-01-2021
           """

__all__ = ["duplicate_columns", "ExportMethodEnum", "ChainedAssignmentOptionEnum"]

import enum
from typing import List

import pandas
from pandas.core.dtypes.missing import array_equivalent
from sorcery import assigned_names


class ExportMethodEnum(enum.Enum):
    """
    Available Pandas Dataframe Export methods"""

    (
        parquet,
        pickle,  # also 'dataframe'
        csv,
        hdf,
        sql,
        dict,
        excel,
        json,
        html,
        feather,
        latex,
        stata,
        gbq,
        records,
        string,
        clipboard,
        markdown,
        xarray,
    ) = assigned_names()


class ChainedAssignmentOptionEnum(enum.Enum):
    """
    from contextlib import suppress
    from pandas.core.common import SettingWithCopyWarning
    """

    (
        warn,  # the default, means a SettingWithCopyWarning is printed.
        raises,  # means pandas will raise a SettingWithCopyException you have to deal with.
    ) = assigned_names()
    none = None  # will suppress the warnings entirely.


def duplicate_columns(frame: pandas.DataFrame) -> List[str]:
    """ """
    groups = frame.columns.to_series().groupby(frame.dtypes).groups
    duplicates = []

    for t, v in groups.items():

        cs = frame[v].columns
        vs = frame[v]
        lcs = len(cs)

        for i in range(lcs):
            ia = vs.iloc[:, i].values
            for j in range(i + 1, lcs):
                ja = vs.iloc[:, j].values
                if array_equivalent(ia, ja):
                    duplicates.append(cs[i])
                    break

    return duplicates


if __name__ == "__main__":
    for e in ExportMethodEnum:
        print(e.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/10/2020
           """

__all__ = []

from draugr.os_utilities.linux_utilities.gtk_utilities import GtkThemePreferences


def main() -> None:
    """
    :rtype: None
    """
    with GtkThemePreferences() as a:
        a.prefer_dark_mode = not a.prefer_dark_mode


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04-04-2021
           """

import warnings
from typing import Any

import numpy
from skimage import color, img_as_ubyte

__all__ = ["rgb_to_grayscale"]

from draugr.numpy_utilities.mixing import mix_channels


def rgb_to_grayscale(obs: Any) -> Any:
    """Convert a 3-channel color observation image to grayscale and uint8.

    Args:
       obs (numpy.ndarray): Observation array, conforming to observation_space

    Returns:
       numpy.ndarray: 1-channel grayscale version of obs, represented as uint8

    """
    with warnings.catch_warnings():
        # Suppressing warning for possible precision loss when converting
        # from float64 to uint8
        warnings.simplefilter("ignore")
        return img_as_ubyte(color.rgb2gray(obs))


if __name__ == "__main__":

    def asuijhd():
        a = numpy.expand_dims(numpy.eye(5), -1)
        b = numpy.expand_dims(numpy.flip(numpy.eye(5), 0), -1)
        c = numpy.expand_dims(numpy.ones((5, 5)), -1)
        stacked = numpy.swapaxes(numpy.array((a, b, c)), 0, -1)
        print(stacked.shape)
        print(mix_channels(stacked))

    asuijhd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Iterable, Sequence, Union

import numpy
import torch

__author__ = "Christian Heider Nielsen"
__doc__ = ""
__all__ = ["to_scalar"]

from draugr.torch_utilities.tensors.to_tensor import to_tensor
from warg import Number


def to_scalar(
    obj: Union[torch.Tensor, numpy.ndarray, Iterable, Sequence, int, float],
    device: Union[str, torch.device] = "cpu",
    aggregation: callable = torch.mean,
) -> Number:
    """
    Always detaches from computation graph

    default behaviour is obj.cpu().mean().item()

    :param aggregation:
    :param obj:
    :param device:
    :return:"""

    if not torch.is_tensor(obj):
        obj = to_tensor(obj)
    else:
        obj = obj.detach()

    if aggregation:
        obj = aggregation(obj)

    return obj.to(device=device).item()


if __name__ == "__main__":
    print(to_scalar(0))
    print(to_scalar(2.5))
    print(to_scalar(to_tensor(1).item()))
    print(to_scalar(to_tensor(1)))
    print(to_scalar(to_tensor(2.0)))

    print(to_scalar(to_tensor([0.5, 0.5])))
    print(to_scalar(to_tensor([[0.5, 0.5]])))
    print(to_scalar(to_tensor((0.5, 0.5))))
    print(to_scalar(to_tensor(range(10))))
    print(to_scalar(to_tensor(torch.from_numpy(numpy.array([0.5, 0.5])))))
    a = torch.arange(0, 10)
    print(to_scalar(to_tensor(a)))
    print(to_scalar(to_tensor([a, a])))
    print(to_scalar(to_tensor((a, a))))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09-02-2021
           """

import contextlib
import signal

__all__ = ["IgnoreInterruptSignal"]


class IgnoreInterruptSignal(contextlib.AbstractContextManager):
    def __enter__(self) -> bool:
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        # signal.getsignal() No sideeffect options for exit
        return True

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        signal.signal(signal.SIGINT, signal.SIG_DFL)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy
import visdom

from draugr.writers.writer import Writer

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 9/14/19
           """

__all__ = ["VisdomWriter"]

# Visualisation
USE_VISDOM = False
START_VISDOM_SERVER = False
VISDOM_SERVER = "http://localhost"
if not START_VISDOM_SERVER:
    # noinspection PyRedeclaration
    VISDOM_SERVER = "http://visdom.ml"


class VisdomWriter(Writer):
    """ """

    def __init__(self):
        super().__init__()

    def _scalar(self, tag: str, value: float, step: int):
        self.server.line(
            Y=numpy.array([value]), X=numpy.array([step]), win=tag, update="append"
        )

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        # self.server.close() # close a window by id
        del self.server

    def _open(self):
        self.server = visdom.Visdom(server=VISDOM_SERVER, port=8097)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 31-03-2021
           """

import cv2
import numpy

from warg import DoubleNumber, Number, TripleNumber

__all__ = ["blit_numbering_raster_sequence", "blit_fps"]


def blit_numbering_raster_sequence(
    seq: numpy.ndarray,
    *,
    placement: DoubleNumber = (0, 30),
    color: TripleNumber = (
        0,
        0,
        1,
    ),  # If your images are in [0, 255] range replace (0, 0, 1) with (0, 0, 255)
    thickness: Number = 2,
    font_scale: Number = 1,
    font: int = cv2.FONT_HERSHEY_COMPLEX,
) -> numpy.ndarray:
    """

    :param seq:
    :param placement:
    :param color:
    :param thickness:
    :param font_scale:
    :param font:
    :return:
    """
    result = numpy.empty_like(seq)
    n = len(seq)
    for i in range(n):
        result[i] = cv2.putText(
            seq[i], f"{i}/{n}", placement, font, font_scale, color, thickness
        )
    return result


def blit_fps(
    seq: numpy.ndarray,
    fps: Number,
    *,
    placement: DoubleNumber = (-140, -10),  # bottom-right corner
    format_str: str = "{0} fps",
    color: TripleNumber = (
        0,
        0,
        1,
    ),  # If your images are in [0, 255] range replace (0, 0, 1) with (0, 0, 255)
    thickness: Number = 2,
    font_scale: Number = 1,
    font: int = cv2.FONT_HERSHEY_COMPLEX,
) -> numpy.ndarray:
    """

    :param seq:
    :param fps:
    :param placement:
    :param format_str:
    :param color:
    :param thickness:
    :param font_scale:
    :param font:
    :return:
    """
    result = numpy.empty_like(seq)
    n = len(seq)
    if placement[0] < 0:
        placement = (result.shape[-2] + placement[0], placement[1])
    if placement[1] < 0:
        placement = (placement[0], result.shape[-3] + placement[1])

    for i in range(n):
        result[i] = cv2.putText(
            seq[i],
            format_str.format(fps),
            placement,
            font,
            font_scale,
            color,
            thickness,
        )
    return result


if __name__ == "__main__":

    def asd7ad() -> None:
        """
        :rtype: None
        """
        from pathlib import Path
        from apppath import ensure_existence
        from matplotlib import pyplot
        import numpy
        import imageio

        n = 200
        n_frames = 25
        x = numpy.linspace(-numpy.pi * 4, numpy.pi * 4, n)
        base = ensure_existence(Path("exclude"))

        def gen():
            """ """
            for i, t in enumerate(numpy.linspace(0, numpy.pi, n_frames)):
                pyplot.plot(x, numpy.cos(x + t))
                pyplot.plot(x, numpy.sin(2 * x - t))
                pyplot.plot(x, numpy.cos(x + t) + numpy.sin(2 * x - t))
                pyplot.ylim(-2.5, 2.5)
                pyplot.savefig(base / f"frame{i}.png", bbox_inches="tight", dpi=300)
                pyplot.clf()

        def asijsd():
            """ """
            files = [base / f"frame{yu}.png" for yu in range(n_frames)]
            frames = [imageio.imread(f) for f in files]
            frames = blit_numbering_raster_sequence(frames)
            imageio.mimsave(base / "output.gif", frames, fps=(n_frames / 2.0))

        def sadasf():
            """ """
            files = [base / f"frame{yu}.png" for yu in range(n_frames)]
            a = [imageio.imread(f) for f in files]
            frames = numpy.array([a, a])  # copy of itself, just for test
            fps = n_frames / 2.0
            frames = numpy.array(
                [blit_fps(blit_numbering_raster_sequence(f), fps) for f in frames]
            )
            [
                imageio.mimsave(base / f"output{i}.gif", f, fps=fps)
                for i, f in enumerate(frames)
            ]

        gen()
        sadasf()

    asd7ad()



from typing import Any

from matplotlib import pyplot

__all__ = ["array_to_color"]


def array_to_color(array: Any, cmap: str = "Oranges") -> object:
    """
    Args:
      array:
      cmap:

    Returns:

    """
    return pyplot.cm.ScalarMappable(cmap=cmap).to_rgba(array)[:, :-1]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import time
from typing import List, Sequence

import torch
from torch import nn
from torch.distributions import MultivariateNormal, Normal

from draugr.torch_utilities.optimisation.parameters.initialisation import fan_in_init
from draugr.torch_utilities.architectures.mlp import MLP

__author__ = "Christian Heider Nielsen"
__doc__ = ""
__all__ = [
    "ShallowStdNormalMLP",
    "MultiDimensionalNormalMLP",
    "MultiVariateNormalMLP",
    "MultipleNormalMLP",
]

from warg import passes_kws_to


class ShallowStdNormalMLP(MLP):
    def __init__(
        self,
        output_shape: Sequence = (2,),
        mean_head_activation: callable = None,
        fixed_log_std: torch.Tensor = None,
        **kwargs
    ):
        super().__init__(
            output_shape=output_shape,
            mean_head_activation=mean_head_activation,
            **kwargs
        )

        self.mean_head_activation = mean_head_activation
        if fixed_log_std is None:
            self.log_std = nn.Parameter(
                torch.zeros(output_shape[-1]), requires_grad=True
            )
        else:
            assert fixed_log_std == output_shape[-1]
            self.log_std = fixed_log_std

    def forward(
        self, *x, min_std=-20, max_std=2, **kwargs
    ) -> torch.distributions.Distribution:
        mean = super().forward(*x, min_std=min_std, **kwargs)
        if self.mean_head_activation:
            mean = self.mean_head_activation(mean)

        log_std = torch.clamp(self.log_std, min_std, max_std)
        std = log_std.exp().expand_as(mean)
        return Normal(mean, std)


class ShallowStdMultiVariateNormalMLP(MLP):
    def __init__(
        self,
        output_shape: Sequence = (2,),
        mean_head_activation: callable = None,
        fixed_log_std: torch.Tensor = None,
        **kwargs
    ):
        super().__init__(
            output_shape=output_shape,
            mean_head_activation=mean_head_activation,
            **kwargs
        )

        self.mean_head_activation = mean_head_activation
        if fixed_log_std is None:
            self.log_std = nn.Parameter(torch.zeros(*output_shape), requires_grad=True)
        else:
            assert fixed_log_std.shape == output_shape
            self.log_std = fixed_log_std

    def forward(self, *x, min_std=-20, max_std=2, **kwargs):
        mean = super().forward(*x, min_std=min_std, **kwargs)
        if self.mean_head_activation:
            mean = self.mean_head_activation(mean)

        log_std = torch.clamp(self.log_std, min_std, max_std)
        std = log_std.exp().expand_as(mean)
        std = torch.diag_embed(std, 0, dim1=-2, dim2=-1)
        return MultivariateNormal(mean, std)


class MultiDimensionalNormalMLP(MLP):
    def __init__(
        self,
        output_shape: Sequence = (2,),
        mean_head_activation: callable = None,
        **kwargs
    ):
        output_shape = (*output_shape, *output_shape)
        assert len(output_shape) == 2

        self.mean_head_activation = mean_head_activation
        super().__init__(
            output_shape=output_shape,
            mean_head_activation=mean_head_activation,
            **kwargs
        )

    def forward(self, *x, min_std=-20, max_std=2, **kwargs) -> Normal:
        mean, log_std = super().forward(*x, min_std=min_std, **kwargs)
        if self.mean_head_activation:
            mean = self.mean_head_activation(mean)

        return Normal(mean, torch.clamp(log_std, min_std, max_std).exp())


class MultiVariateNormalMLP(MLP):
    @passes_kws_to(MLP.__init__)
    def __init__(
        self,
        output_shape: Sequence = (2,),
        mean_head_activation: callable = None,
        **kwargs
    ):
        output_shape = (*output_shape, *output_shape)
        assert len(output_shape) == 2

        self.mean_head_activation = mean_head_activation
        super().__init__(output_shape=output_shape, **kwargs)

    @passes_kws_to(MLP.forward)
    def forward(self, *x, min_std=-20, max_std=2, **kwargs) -> MultivariateNormal:
        mean, log_std = super().forward(*x, min_std=min_std, **kwargs)
        if self.mean_head_activation:
            mean = self.mean_head_activation(mean)

        log_std = torch.clamp(log_std, min_std, max_std)
        std = log_std.exp()
        std = torch.diag_embed(std, 0, dim1=-2, dim2=-1)
        return MultivariateNormal(mean, std)


class MultipleNormalMLP(MLP):
    def __init__(
        self, output_shape: int = 2, mean_head_activation: callable = None, **kwargs
    ):
        output_shape = (2,) * output_shape
        super().__init__(output_shape=output_shape, **kwargs)

        self.mean_head_activation = mean_head_activation
        fan_in_init(self)

    def forward(self, *x, min_std=-20, max_std=2, **kwargs) -> List[Normal]:
        out = super().forward(*x, min_std=min_std, **kwargs)
        outs = []
        for mean, log_std in out:
            if self.mean_head_activation:
                mean = self.mean_head_activation(mean)
            outs.append(Normal(mean, torch.clamp(log_std, min_std, max_std).exp()))

        return outs


if __name__ == "__main__":

    def stest_normal():
        s = (10,)
        a = 10
        model = MultipleNormalMLP(input_shape=s, output_shape=a)

        inp = torch.rand(s)
        s_ = time.time()
        dis = model.forward(inp)
        print(dis)
        a_ = [d.sample() for d in dis]
        print(time.time() - s_, a_)

    def stest_multi_dim_normal():
        s = (4,)
        a = (10,)
        model = MultiDimensionalNormalMLP(input_shape=s, output_shape=a)

        inp = torch.rand(s)
        s_ = time.time()
        dis = model.forward(inp)
        print(dis)
        a_ = dis.sample()
        print(time.time() - s_, a_)

    def stest_multi_var_normal():
        s = (10,)
        a = (10,)
        model = MultiVariateNormalMLP(input_shape=s, output_shape=a)

        inp = torch.rand(s)
        s_ = time.time()
        dis = model.forward(inp)
        print(dis)
        a_ = dis.sample()
        print(time.time() - s_, a_)

    def stest_shallow():
        s = (10,)
        a = (10,)
        model = ShallowStdNormalMLP(input_shape=s, output_shape=a)

        inp = torch.rand(s)
        s_ = time.time()
        dis = model.forward(inp)
        print(dis)
        a_ = dis.sample()
        print(time.time() - s_, a_)

    stest_normal()
    print("\n")
    stest_multi_dim_normal()
    print("\n")
    stest_multi_var_normal()
    print("\n")
    stest_shallow()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/07/2020
           """

__all__ = ["save_config", "ensure_directory_exist"]

import shutil

from pathlib import Path


def save_config(config_save_path: Path, config_file_path: Path) -> None:
    """

    :param config_save_path:
    :param config_file_path:
    :return:"""
    shutil.copyfile(str(config_file_path), str(config_save_path))


def ensure_directory_exist(model_path: Path) -> None:
    """

    :param model_path:
    :return:"""
    if not model_path.exists():
        model_path.mkdir(parents=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

import numpy
import torch
from torch.nn import Conv2d, Linear, Module
from torch.nn.init import calculate_gain, constant_, uniform_, xavier_uniform_

__all__ = ["fan_in_init", "xavier_init", "constant_init", "normal_init"]


def fan_in_init(model: Module):
    """

    :param model:
    :type model:"""
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            fan_in = m.weight.size(1)
            v = 1.0 / numpy.sqrt(fan_in)
            uniform_(m.weight, -v, v)
            constant_(m.bias, 0)


def xavier_init(model: Module, activation="relu"):
    """

    :param model:
    :type model:
    :param activation:
    :type activation:"""
    gain = calculate_gain(activation)
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            xavier_uniform_(m.weight, gain=gain)
            constant_(m.bias, 0)


def constant_init(model: Module, constant: float = 1):
    """

    :param model:
    :type model:
    :param constant:
    :type constant:"""
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            constant_(m.weight, constant)
            constant_(m.bias, constant)


def normal_init(model: Module, mean: float = 0, std: float = 1.0):
    """

    :param model:
    :param mean:
    :param std:
    :return:
    """
    for m in model.modules():
        if isinstance(m, (Conv2d, Linear)):
            torch.nn.init.normal_(m.weight, mean, std)
            torch.nn.init.normal_(m.bias, mean, std)


if __name__ == "__main__":
    a = Linear(3, 3)
    fan_in_init(a)
    xavier_init(a)
    constant_init(a)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["InstantiationWriterMixin"]

from typing import Mapping

from warg import drop_unused_kws


class InstantiationWriterMixin(ABC):
    """
    Writer mixin that provides an interface for 'writing' instantiation"""

    @drop_unused_kws
    @abstractmethod
    def instance(self, instance: Mapping, metrics: Mapping) -> None:
        """ """
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

import torch

__all__ = [
    "BatchToTensor",
    "BatchNormalize",
    "BatchRandomCrop",
    "BatchRandomHorizontalFlip",
]


class BatchToTensor:
    """Applies the :class:`~torchvision.transforms.ToTensor` transform to a batch of images."""

    def __init__(self):
        self.max = 255

    def __call__(self, tensor):
        """
        Args:
        tensor (Tensor): Tensor of size (N, C, H, W) to be tensorized.
        Returns:
        Tensor: Tensorized Tensor."""
        return tensor.float().div_(self.max)


class BatchNormalize:
    """Applies the :class:`~torchvision.transforms.Normalize` transform to a batch of images.
    .. note::
    This transform acts out of place by default, i.e., it does not mutate the input tensor.
    Args:
    mean (sequence): Sequence of means for each channel.
    std (sequence): Sequence of standard deviations for each channel.
    inplace(bool,optional): Bool to make this operation in-place.
    dtype (torch.dtype,optional): The data type of tensors to which the transform will be applied.
    device (torch.device,optional): The device of tensors to which the transform will be applied."""

    def __init__(self, mean, std, inplace=False, dtype=torch.float, device="cpu"):
        self.mean = torch.as_tensor(mean, dtype=dtype, device=device)[
            None, :, None, None
        ]
        self.std = torch.as_tensor(std, dtype=dtype, device=device)[None, :, None, None]
        self.inplace = inplace

    def __call__(self, tensor):
        """
        Args:
        tensor (Tensor): Tensor of size (N, C, H, W) to be normalized.
        Returns:
        Tensor: Normalized Tensor."""
        if not self.inplace:
            tensor = tensor.clone()

        tensor.sub_(self.mean).div_(self.std)
        return tensor


class BatchRandomHorizontalFlip:
    """Applies the :class:`~torchvision.transforms.RandomHorizontalFlip` transform to a batch of images.
    .. note::
    This transform acts out of place by default, i.e., it does not mutate the input tensor.
    Args:
    p (float): probability of an image being flipped.
    inplace(bool,optional): Bool to make this operation in-place."""

    def __init__(self, p=0.5, inplace=False):
        self.p = p
        self.inplace = inplace

    def __call__(self, tensor):
        """
        Args:
        tensor (Tensor): Tensor of size (N, C, H, W) to be flipped.
        Returns:
        Tensor: Randomly flipped Tensor."""
        if not self.inplace:
            tensor = tensor.clone()

        flipped = torch.rand(tensor.size(0)) < self.p
        tensor[flipped] = torch.flip(tensor[flipped], [3])
        return tensor


class BatchRandomCrop:
    """Applies the :class:`~torchvision.transforms.RandomCrop` transform to a batch of images.
    Args:
    size (int): Desired output size of the crop.
    padding (int, optional): Optional padding on each border of the image.
        Default is None, i.e no padding.
    dtype (torch.dtype,optional): The data type of tensors to which the transform will be applied.
    device (torch.device,optional): The device of tensors to which the transform will be applied."""

    def __init__(self, size, padding=None, dtype=torch.float, device="cpu"):
        self.size = size
        self.padding = padding
        self.dtype = dtype
        self.device = device

    def __call__(self, tensor):
        """
        Args:
        tensor (Tensor): Tensor of size (N, C, H, W) to be cropped.
        Returns:
        Tensor: Randomly cropped Tensor."""
        if self.padding is not None:
            padded = torch.zeros(
                (
                    tensor.size(0),
                    tensor.size(1),
                    tensor.size(2) + self.padding * 2,
                    tensor.size(3) + self.padding * 2,
                ),
                dtype=self.dtype,
                device=self.device,
            )
            padded[
                :, :, self.padding : -self.padding, self.padding : -self.padding
            ] = tensor
        else:
            padded = tensor

        w, h = padded.size(2), padded.size(3)
        th, tw = self.size, self.size
        if w == tw and h == th:
            i, j = 0, 0
        else:
            i = torch.randint(0, h - th + 1, (tensor.size(0),), device=self.device)
            j = torch.randint(0, w - tw + 1, (tensor.size(0),), device=self.device)

        rows = torch.arange(th, dtype=torch.long, device=self.device) + i[:, None]
        columns = torch.arange(tw, dtype=torch.long, device=self.device) + j[:, None]
        padded = padded.permute(1, 0, 2, 3)
        padded = padded[
            :,
            torch.arange(tensor.size(0))[:, None, None],
            rows[:, torch.arange(th)[:, None]],
            columns[:, None],
        ]
        return padded.permute(1, 0, 2, 3)
